from .base_box import BaseBox
from .base_cone import BaseCone
from .base_cylinder import BaseCylinder
from .base_lens import BaseLens
from .base_logo_box import BaseLogoBox
from .base_pipe import BasePipe
from .base_pipe_segment import BasePipeSegment
from .base_revolve import BaseRevolve
from .base_ring import BaseRing
from .base_shape import BaseShape
from .base_sphere import BaseSphere
from .lattice_manifold import LatticeManifold
from .lattice_pipe import LatticePipe

__all__ = [
    "BaseShape",
    "BaseBox",
    "BaseCylinder",
    "BasePipe",
    "BasePipeSegment",
    "BaseSphere",
    "BaseLens",
    "BaseRing",
    "BaseCone",
    "BaseRevolve",
    "LatticePipe",
    "LatticeManifold",
    "BaseLogoBox",
]



from __future__ import annotations

from collections.abc import Callable


class BisectionException(Exception):
	pass


class Bisection:
	Func = Callable[[float], float]

	def __init__(
		self,
		oFunc: Func,
		fMinInput: float,
		fMaxInput: float,
		fTargetOutput: float,
		fEpsilon: float = 0.01,
		nMaxIterations: int = 500,
	) -> None:
		self.m_oFunc = oFunc
		self.m_fEpsilon = float(fEpsilon)
		self.m_fMinInput = float(fMinInput)
		self.m_fMaxInput = float(fMaxInput)
		self.m_fTargetOutput = float(fTargetOutput)
		self.m_nIterations = 0
		self.m_nMaxIterations = int(nMaxIterations)
		self.m_fRemainingDiff = float(self.m_fMaxInput - self.m_fMinInput)
		self.m_fBestGuess = float(self.m_fMinInput)

	def fGetOutputFromFunc(self, fInput: float) -> float:
		return float(self.m_oFunc(float(fInput))) - self.m_fTargetOutput

	def fFindOptimalInput(self) -> float:
		f_min = self.m_fMinInput
		f_max = self.m_fMaxInput
		out_min = self.fGetOutputFromFunc(f_min)
		out_max = self.fGetOutputFromFunc(f_max)
		if out_min * out_max >= 0.0:
			raise BisectionException("No valid limits.")

		mid = f_min
		self.m_fRemainingDiff = f_max - f_min
		while self.m_fRemainingDiff >= self.m_fEpsilon:
			mid = 0.5 * (f_min + f_max)
			out_mid = self.fGetOutputFromFunc(mid)
			if out_mid == 0.0:
				break
			if out_mid * self.fGetOutputFromFunc(f_min) < 0.0:
				f_max = mid
			else:
				f_min = mid

			self.m_fRemainingDiff = f_max - f_min
			self.m_nIterations += 1
			self.m_fBestGuess = mid

			if self.m_nIterations == self.m_nMaxIterations:
				raise BisectionException("No solution reached after max number of interations.")

		return float(mid)

	def nGetIterations(self) -> int:
		return int(self.m_nIterations)

	def fGetRemainingDiff(self) -> float:
		return float(self.m_fRemainingDiff)

	def fGetBestGuess(self) -> float:
		return float(self.m_fBestGuess)


__all__ = ["Bisection", "BisectionException"]



from __future__ import annotations

import math

from shape_kernel._types import Vector3Like, as_vec3
from shape_kernel.utils import VecOperations

from picogk import IImplicit


class ImplicitRadialGyroid(IImplicit):
	def __init__(self, nUnitsPerRound: int, fUnitSizeInZ: float, fWallThickness: float) -> None:
		self.m_fFrequencyScale = (2.0 * math.pi) / float(fUnitSizeInZ)
		self.m_fWallThickness = float(fWallThickness)
		self.m_nSamplesPerRound = int(nUnitsPerRound)

	def fSignedDistance(self, vecPt: Vector3Like) -> float:
		pt = as_vec3(vecPt)
		fRadius = VecOperations.fGetRadius(pt)
		dPhi = (2.0 * math.pi) / float(self.m_nSamplesPerRound)
		fPhi = VecOperations.fGetPhi(pt) + math.pi
		fPhiIntervals = fPhi / dPhi
		fUnitSize = (2.0 * math.pi) / self.m_fFrequencyScale
		dX = fRadius
		dY = fPhiIntervals * fUnitSize
		dZ = float(pt[2])

		dDist = (
			math.sin(self.m_fFrequencyScale * dX) * math.cos(self.m_fFrequencyScale * dY)
			+ math.sin(self.m_fFrequencyScale * dY) * math.cos(self.m_fFrequencyScale * dZ)
			+ math.sin(self.m_fFrequencyScale * dZ) * math.cos(self.m_fFrequencyScale * dX)
		)
		return abs(dDist) - 0.5 * self.m_fWallThickness


__all__ = ["ImplicitRadialGyroid"]


from __future__ import annotations

from collections.abc import Sequence

from .._types import Vec3, Vector3Like, as_vec3


class GridOperations:
	@staticmethod
	def aGetListInX(aGrid: Sequence[Sequence[Vector3Like]], index: int) -> list[Vec3]:
		return [as_vec3(p) for p in aGrid[int(index)]]

	@staticmethod
	def aGetListInY(aGrid: Sequence[Sequence[Vector3Like]], index: int) -> list[Vec3]:
		idx = int(index)
		return [as_vec3(row[idx]) for row in aGrid]

	@staticmethod
	def aAddListInX(aGrid: Sequence[Sequence[Vector3Like]], aList: Sequence[Vector3Like]) -> list[list[Vec3]]:
		out = [[as_vec3(p) for p in row] for row in aGrid]
		out.append([as_vec3(p) for p in aList])
		return out

	@staticmethod
	def aAddListInY(aGrid: Sequence[Sequence[Vector3Like]], aList: Sequence[Vector3Like]) -> list[list[Vec3]]:
		out = [[as_vec3(p) for p in row] for row in aGrid]
		col = [as_vec3(p) for p in aList]
		if len(out) != len(col):
			raise ValueError("Column length must match grid height")
		for i in range(len(out)):
			out[i].append(col[i])
		return out

	@staticmethod
	def aRemoveListInX(aGrid: Sequence[Sequence[Vector3Like]], index: int) -> list[list[Vec3]]:
		idx = int(index)
		out = [[as_vec3(p) for p in row] for row in aGrid]
		if idx < 0 or idx >= len(out):
			raise IndexError(idx)
		del out[idx]
		return out

	@staticmethod
	def aRemoveListInY(aGrid: Sequence[Sequence[Vector3Like]], index: int) -> list[list[Vec3]]:
		idx = int(index)
		out = [[as_vec3(p) for p in row] for row in aGrid]
		for row in out:
			if idx < 0 or idx >= len(row):
				raise IndexError(idx)
			del row[idx]
		return out

	@staticmethod
	def aGetInverseGrid(aGrid: Sequence[Sequence[Vector3Like]]) -> list[list[Vec3]]:
		if not aGrid:
			return []
		h = len(aGrid)
		w = len(aGrid[0])
		return [[as_vec3(aGrid[i][j]) for i in range(h)] for j in range(w)]


__all__ = ["GridOperations"]



from __future__ import annotations

import math
import numpy as np

from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Sequence, cast

from picogk import BBox3, IImplicit, Lattice, Library, Mesh, PolyLine, VedoViewer, Voxels
from picogk._types import ColorFloat

from .._types import ColorLike, Vec3, Vector3Like, as_np3, as_vec3
from ..frames.local_frames import LocalFrame
from ..utils import GridOperations, VecOperations

if TYPE_CHECKING:
    from ..base_shapes import BaseBox, BaseCylinder, BasePipe, BaseRing
    from ..frames import Frames


class Sh:
    nNumberOfGroups = 0

    @staticmethod
    def _resolve_viewer(oViewer: VedoViewer | None = None) -> VedoViewer:
        viewer = oViewer if oViewer is not None else Library.oViewer()
        if viewer is None:
            raise RuntimeError("No active viewer. Pass viewer=... to picogk.go(...) or call the preview helper with oViewer=...")
        return viewer # pyright: ignore[reportReturnType]

    @staticmethod
    def _color_rgba(clrColor: ColorLike, alpha: float | None = None) -> tuple[float, float, float, float]:
        if isinstance(clrColor, ColorFloat):
            rgba = (float(clrColor.r), float(clrColor.g), float(clrColor.b), float(clrColor.a))
        else:
            clr = tuple(float(v) for v in clrColor)
            if len(clr) == 3:
                rgba = (clr[0], clr[1], clr[2], 1.0)
            elif len(clr) == 4:
                rgba = (clr[0], clr[1], clr[2], clr[3])
            else:
                raise ValueError("Color must have 3 or 4 components")
        if alpha is None:
            return rgba
        return (rgba[0], rgba[1], rgba[2], float(alpha))

    @staticmethod
    def _next_group_id() -> int:
        group_id = int(Sh.nNumberOfGroups)
        Sh.nNumberOfGroups += 1
        return group_id

    @staticmethod
    def ResetPreviewGroups() -> None:
        Sh.nNumberOfGroups = 0

    @staticmethod
    def _is_point_like(value: object) -> bool:
        if isinstance(value, (str, bytes)):
            return False
        try:
            seq = cast(Sequence[float], value)
            if len(seq) != 3:
                return False
            float(seq[0])
            float(seq[1])
            float(seq[2])
            return True
        except Exception:
            return False

    @staticmethod
    def _is_point_sequence(value: object) -> bool:
        if isinstance(value, (str, bytes)):
            return False
        try:
            seq = cast(Sequence[object], value)
        except Exception:
            return False
        return len(seq) > 0 and Sh._is_point_like(seq[0])

    @staticmethod
    def _parse_preview_style(args: Sequence[object], start: int) -> tuple[float, float, float]:
        values = [0.9, 0.4, 0.7]
        for index in range(min(3, max(0, len(args) - start))):
            values[index] = float(cast(float, args[start + index]))
        return (values[0], values[1], values[2])

    @staticmethod
    def _preview_curve_samples(
        fnPoint: Callable[[float, float], Vec3],
        clrColor: ColorLike,
        param1_values: Sequence[float],
        param2_values: Sequence[float],
        *,
        nCurveSamples: int,
        oViewer: VedoViewer | None = None,
    ) -> list[int]:
        groups: list[int] = []
        for value in param1_values:
            points = [fnPoint(value, t) for t in param2_values]
            groups.append(Sh.PreviewLine(points, clrColor, oViewer=oViewer))
        if nCurveSamples > 0:
            for value in param2_values:
                points = [fnPoint(t, value) for t in np.linspace(0.0, 1.0, nCurveSamples)]
                groups.append(Sh.PreviewLine(points, clrColor, oViewer=oViewer))
        return groups

    @staticmethod
    def voxOffset(vox: Voxels, fDistInMM: float) -> Voxels:
        return vox.voxOffset(fDistInMM)

    @staticmethod
    def voxSmoothen(vox: Voxels, fSmoothInMM: float) -> Voxels:
        return vox.voxSmoothen(fSmoothInMM)

    @staticmethod
    def voxOverOffset(vox: Voxels, fInitialDistInMM: float, fFinalDistInMM: float) -> Voxels:
        return vox.voxOverOffset(fInitialDistInMM, fFinalDistInMM)

    @staticmethod
    def voxExtrudeZSlice(vox: Voxels, fStartHeight: float, fEndHeight: float) -> Voxels:
        return vox.voxProjectZSlice(fStartHeight, fEndHeight)

    @staticmethod
    def voxUnion(vox1: Voxels | Sequence[Voxels], vox2: Voxels | None = None) -> Voxels:
        if vox2 is None:
            if not isinstance(vox1, Sequence):
                raise TypeError("voxUnion expects (Voxels, Voxels) or (Sequence[Voxels])")
            return Voxels.voxCombineAll(vox1)
        if isinstance(vox1, Sequence):
            raise TypeError("voxUnion binary form expects two Voxels arguments")
        return vox1.voxBoolAdd(vox2)

    @staticmethod
    def voxSubtract(vox1: Voxels, vox2: Voxels) -> Voxels:
        return vox1.voxBoolSubtract(vox2)

    @staticmethod
    def voxIntersect(vox1: Voxels, vox2: Voxels) -> Voxels:
        return vox1.voxBoolIntersect(vox2)

    @staticmethod
    def voxIntersectImplicit(vox: Voxels, sdfObject: IImplicit) -> Voxels:
        return vox.voxIntersectImplicit(sdfObject)

    @staticmethod
    def vecGetClosestSurfacePoint(vox: Voxels, vecPt: Vector3Like) -> tuple[float, float, float]:
        return vox.vecClosestPointOnSurface(vecPt)

    @staticmethod
    def vecGetProjectedSurfacePoint(vox: Voxels, vecPt: Vector3Like, vecDir: Vector3Like) -> tuple[float, float, float]:
        return vox.vecRayCastToSurface(vecPt, vecDir)

    @staticmethod
    def oGetBoundingBox(vox: Voxels) -> tuple[tuple[float, float, float], tuple[float, float, float]]:
        return vox.oCalculateBoundingBox()

    @staticmethod
    def latFromLine(aPoints: Sequence[Vector3Like], fBeam: float) -> Lattice:
        oLattice = Lattice()
        for i in range(1, len(aPoints)):
            oLattice.AddBeam(aPoints[i - 1], aPoints[i], fBeam, fBeam, True)
        return oLattice

    @staticmethod
    def AddLine(oLattice: Lattice, aPoints: Sequence[Vector3Like], fBeam: float) -> None:
        for i in range(1, len(aPoints)):
            oLattice.AddBeam(aPoints[i - 1], aPoints[i], fBeam, fBeam, True)

    @staticmethod
    def latFromPoints(aPoints: Sequence[Vector3Like], fBeam: float) -> Lattice:
        oLattice = Lattice()
        for p in aPoints:
            oLattice.AddSphere(p, fBeam)
        return oLattice

    @staticmethod
    def latFromEdges(aEdges: Sequence[Sequence[Vector3Like]], fBeam: float) -> Lattice:
        oLattice = Lattice()
        for edge in aEdges:
            for i in range(1, len(edge)):
                oLattice.AddBeam(edge[i - 1], edge[i], fBeam, fBeam, True)
        return oLattice

    @staticmethod
    def latFromPoint(vecPt: Vector3Like, fBeam: float) -> Lattice:
        oLattice = Lattice()
        oLattice.AddSphere(vecPt, fBeam)
        return oLattice

    @staticmethod
    def latFromGrid(aGrid: Sequence[Sequence[Vector3Like]], fBeam: float) -> Lattice:
        oLattice = Lattice()
        for row in aGrid:
            Sh.AddLine(oLattice, row, fBeam)
        if aGrid:
            rows = len(aGrid)
            cols = len(aGrid[0])
            for c in range(cols):
                for r in range(1, rows):
                    oLattice.AddBeam(aGrid[r - 1][c], aGrid[r][c], fBeam, fBeam, True)
        return oLattice

    @staticmethod
    def latFromBeam(vecPt1: Vector3Like, vecPt2: Vector3Like, fBeam1: float, fBeam2: float | None = None, bRounded: bool = True) -> Lattice:
        oLattice = Lattice()
        beam2 = fBeam1 if fBeam2 is None else fBeam2
        oLattice.AddBeam(vecPt1, vecPt2, fBeam1, beam2, bRounded)
        return oLattice

    @staticmethod
    def voxShell(vox: Voxels, fNegOffset: float, fPosOffset: float, fSmooth: float = 0.0) -> Voxels:
        return vox.voxShell(fNegOffset, fPosOffset, fSmooth)

    @staticmethod
    def ExportMeshToSTLFile(oMesh: Mesh, strFilePath: str) -> None:
        try:
            oMesh.SaveToStlFile(strFilePath)
            Library.Log("STL Export: {0} exported.", strFilePath)
        except Exception as e:
            Library.Log("Could not save STL: {0}", str(e))

    @staticmethod
    def ExportVoxelsToSTLFile(oVoxels: Voxels, strFilePath: str) -> None:
        with Mesh.from_voxels(oVoxels) as mesh:
            Sh.ExportMeshToSTLFile(mesh, strFilePath)

    @staticmethod
    def ExportVoxelsToVDBFile(oVoxels: Voxels, strFilePath: str) -> None:
        try:
            oVoxels.SaveToVdbFile(strFilePath)
            Library.Log("VDB Export: {0} exported.", strFilePath)
        except Exception as e:
            Library.Log("Could not save VDB: {0}", str(e))

    @staticmethod
    def ExportVoxelsToCLIFile(oVoxels: Voxels, strFilePath: str) -> None:
        try:
            oVoxels.SaveToCliFile(strFilePath)
            Library.Log("CLI Export: {0} exported.", strFilePath)
        except Exception as e:
            Library.Log("Could not save CLI: {0}", str(e))

    class EExport(Enum):
        STL = "STL"
        TGA = "TGA"
        PNG = "PNG"
        CSV = "CSV"
        VDB = "VDB"
        CLI = "CLI"

    @staticmethod
    def strGetExportPath(eExport: "Sh.EExport", strFilename: str) -> str:
        ext = f".{eExport.value}"
        return str(Path.cwd() / f"{strFilename}{ext}")

    @staticmethod
    def PreviewMesh(
        oMesh: Mesh,
        clrColor: ColorLike,
        fTransparency: float = 0.9,
        fMetallic: float = 0.4,
        fRoughness: float = 0.7,
        *,
        oViewer: VedoViewer | None = None,
    ) -> int:
        viewer = Sh._resolve_viewer(oViewer)
        group_id = Sh._next_group_id()
        viewer.SetGroupMaterial(group_id, Sh._color_rgba(clrColor, fTransparency), fMetallic, fRoughness)
        viewer.Add(oMesh, group_id)
        return group_id

    @staticmethod
    def PreviewVoxels(
        oVoxels: Voxels,
        clrColor: ColorLike,
        fTransparency: float = 0.9,
        fMetallic: float = 0.4,
        fRoughness: float = 0.7,
        *,
        oViewer: VedoViewer | None = None,
    ) -> int:
        viewer = Sh._resolve_viewer(oViewer)
        group_id = Sh._next_group_id()
        viewer.SetGroupMaterial(group_id, Sh._color_rgba(clrColor, fTransparency), fMetallic, fRoughness)
        viewer.Add(oVoxels, group_id)
        return group_id

    @staticmethod
    def PreviewLattice(
        oLattice: Lattice,
        clrColor: ColorLike,
        fTransparency: float = 0.9,
        fMetallic: float = 0.4,
        fRoughness: float = 0.7,
        *,
        oViewer: VedoViewer | None = None,
    ) -> int:
        with Voxels.from_lattice(oLattice) as oVoxels:
            return Sh.PreviewVoxels(oVoxels, clrColor, fTransparency, fMetallic, fRoughness, oViewer=oViewer)

    @staticmethod
    def PreviewPoint(
        vecPt: Vector3Like,
        fBeam: float,
        clrColor: ColorLike,
        fTransparency: float = 0.9,
        fMetallic: float = 0.4,
        fRoughness: float = 0.7,
        *,
        oViewer: VedoViewer | None = None,
    ) -> int:
        from ..base_shapes import BaseSphere

        oBall = BaseSphere(LocalFrame(vecPt), fBeam)
        return Sh.PreviewMesh(oBall.mshConstruct(), clrColor, fTransparency, fMetallic, fRoughness, oViewer=oViewer)

    @staticmethod
    def PreviewBeam(
        vecPt1: Vector3Like,
        fBeam1: float,
        vecPt2: Vector3Like,
        fBeam2: float,
        clrColor: ColorLike,
        fTransparency: float = 0.9,
        fMetallic: float = 0.4,
        fRoughness: float = 0.7,
        *,
        oViewer: VedoViewer | None = None,
    ) -> int:
        oLattice = Lattice()
        oLattice.AddBeam(vecPt1, vecPt2, fBeam1, fBeam2, True)
        return Sh.PreviewLattice(oLattice, clrColor, fTransparency, fMetallic, fRoughness, oViewer=oViewer)

    @staticmethod
    def PreviewLine(
        aPoints_or_pt1: Sequence[Vector3Like] | Vector3Like,
        clrColor_or_pt2: ColorLike | Vector3Like,
        clrColor: ColorLike | None = None,
        *,
        oViewer: VedoViewer | None = None,
    ) -> int:
        if clrColor is None:
            points = cast(Sequence[Vector3Like], aPoints_or_pt1)
            color = cast(ColorLike, clrColor_or_pt2)
        else:
            points = [cast(Vector3Like, aPoints_or_pt1), cast(Vector3Like, clrColor_or_pt2)]
            color = clrColor

        viewer = Sh._resolve_viewer(oViewer)
        group_id = Sh._next_group_id()
        with PolyLine(Sh._color_rgba(color)) as oLine:
            oLine.Add(points)
            viewer.Add(oLine, group_id)
        return group_id

    @staticmethod
    def PreviewGrid(aGrid: Sequence[Sequence[Vector3Like]], clrColor: ColorLike, *, oViewer: VedoViewer | None = None) -> list[int]:
        groups = [Sh.PreviewLine(row, clrColor, oViewer=oViewer) for row in aGrid]
        groups.extend(Sh.PreviewLine(row, clrColor, oViewer=oViewer) for row in GridOperations.aGetInverseGrid(aGrid))
        return groups

    @staticmethod
    def PreviewEdges(aEdges: Sequence[Sequence[Vector3Like]], clrColor: ColorLike, *, oViewer: VedoViewer | None = None) -> list[int]:
        return [Sh.PreviewLine(edge, clrColor, oViewer=oViewer) for edge in aEdges]

    @staticmethod
    def PreviewPointCloud(
        aPoints: Sequence[Vector3Like],
        fBeam: float,
        clrColor: ColorLike,
        fTransparency: float = 0.9,
        fMetallic: float = 0.4,
        fRoughness: float = 0.7,
        *,
        oViewer: VedoViewer | None = None,
    ) -> list[int]:
        return [
            Sh.PreviewPoint(point, fBeam, clrColor, fTransparency, fMetallic, fRoughness, oViewer=oViewer)
            for point in aPoints
        ]

    @staticmethod
    def PreviewFrame(oFrame: LocalFrame, fSize: float, *, oViewer: VedoViewer | None = None) -> list[int]:
        from ..visualizations import Cp

        viewer = Sh._resolve_viewer(oViewer)
        pos = as_np3(oFrame.vecGetPosition())
        axes = [
            (Cp.clrRed, as_np3(oFrame.vecGetLocalX())),
            (Cp.clrGreen, as_np3(oFrame.vecGetLocalY())),
            (Cp.clrBlue, as_np3(oFrame.vecGetLocalZ())),
        ]
        groups: list[int] = [Sh.PreviewPoint(oFrame.vecGetPosition(), 0.5, Cp.clrBlack, oViewer=viewer)]
        for clr, axis in axes:
            group_id = Sh._next_group_id()
            with PolyLine(Sh._color_rgba(clr)) as oLine:
                oLine.nAddVertex(as_vec3(pos))
                oLine.nAddVertex(as_vec3(pos + float(fSize) * axis))
                oLine.AddArrow(0.2 * float(fSize))
                viewer.Add(oLine, group_id)
            groups.append(group_id)
        return groups

    @staticmethod
    def PreviewFrames(aFrames: Frames, fSize: float, *, oViewer: VedoViewer | None = None) -> list[int]:
        from ..frames import Frames
        from ..utils import SplineOperations

        if not isinstance(aFrames, Frames):
            raise TypeError("PreviewFrames expects a Frames instance")
        fTotalLength = SplineOperations.fGetTotalLength(aFrames.aGetPoints(100))
        nSamples = max(2, int(2.0 * fTotalLength / max(float(fSize), 1e-9)))
        groups: list[int] = []
        for i in range(nSamples):
            fLR = float(i) / float(max(1, nSamples - 1))
            groups.extend(Sh.PreviewFrame(aFrames.oGetLocalFrame(fLR), fSize, oViewer=oViewer))
        return groups

    @staticmethod
    def PreviewCircleSection(
        oFrame: LocalFrame,
        fRadius: float,
        clrColor: ColorLike,
        fTransparency: float = 0.9,
        fMetallic: float = 0.4,
        fRoughness: float = 0.7,
        *,
        oViewer: VedoViewer | None = None,
    ) -> int:
        vecPt1 = as_np3(oFrame.vecGetPosition())
        vecPt2 = vecPt1 + as_np3(oFrame.vecGetLocalZ())
        oLattice = Lattice()
        oLattice.AddBeam(as_vec3(vecPt1), as_vec3(vecPt2), fRadius, fRadius, False)
        return Sh.PreviewLattice(oLattice, clrColor, fTransparency, fMetallic, fRoughness, oViewer=oViewer)

    @staticmethod
    def PreviewCircle(
        oFrame_or_centre: LocalFrame | Vector3Like,
        fRadius: float,
        clrColor: ColorLike,
        *,
        oViewer: VedoViewer | None = None,
    ) -> int:
        if isinstance(oFrame_or_centre, LocalFrame):
            oFrame = oFrame_or_centre
        else:
            oFrame = LocalFrame(oFrame_or_centre)
        aPoints = [
            VecOperations.vecTranslatePointOntoFrame(oFrame, VecOperations.vecGetCylPoint(fRadius, 2.0 * math.pi * i / 99.0, 0.0))
            for i in range(100)
        ]
        return Sh.PreviewLine(aPoints, clrColor, oViewer=oViewer)

    @staticmethod
    def PreviewCylinderWireframe(
        oCyl: BaseCylinder,
        clrColor: ColorLike,
        nRadialSamples: int = 4,
        nLengthSamples: int = 10,
        *,
        oViewer: VedoViewer | None = None,
    ) -> list[int]:
        from ..base_shapes import BaseCylinder

        if not isinstance(oCyl, BaseCylinder):
            raise TypeError("PreviewCylinderWireframe expects a BaseCylinder")
        groups: list[int] = []
        for nLengthSample in range(max(2, int(nLengthSamples))):
            fLengthRatio = float(nLengthSample) / float(max(1, nLengthSamples - 1))
            points = [oCyl.vecGetSurfacePoint(fLengthRatio, float(i) / 99.0, 1.0) for i in range(100)]
            groups.append(Sh.PreviewLine(points, clrColor, oViewer=oViewer))
        for nRadialSample in range(max(1, int(nRadialSamples))):
            fPhiRatio = float(nRadialSample) / float(max(1, nRadialSamples))
            points = [oCyl.vecGetSurfacePoint(float(i) / 999.0, fPhiRatio, 1.0) for i in range(1000)]
            groups.append(Sh.PreviewLine(points, clrColor, oViewer=oViewer))
        return groups

    @staticmethod
    def PreviewPipeWireframe(
        oPipe: BasePipe,
        clrColor: ColorLike,
        nRadialSamples: int = 4,
        nLengthSamples: int = 10,
        *,
        oViewer: VedoViewer | None = None,
    ) -> list[int]:
        from ..base_shapes import BasePipe

        if not isinstance(oPipe, BasePipe):
            raise TypeError("PreviewPipeWireframe expects a BasePipe")
        groups: list[int] = []
        for radius_ratio in (1.0, 0.0):
            for nLengthSample in range(max(2, int(nLengthSamples))):
                fLengthRatio = float(nLengthSample) / float(max(1, nLengthSamples - 1))
                points = [oPipe.vecGetSurfacePoint(fLengthRatio, float(i) / 99.0, radius_ratio) for i in range(100)]
                groups.append(Sh.PreviewLine(points, clrColor, oViewer=oViewer))
        for radius_ratio in (1.0, 0.0):
            for nRadialSample in range(max(2, int(nRadialSamples))):
                fPhiRatio = float(nRadialSample) / float(max(1, nRadialSamples - 1))
                points = [oPipe.vecGetSurfacePoint(float(i) / 99.0, fPhiRatio, radius_ratio) for i in range(100)]
                groups.append(Sh.PreviewLine(points, clrColor, oViewer=oViewer))
        return groups

    @staticmethod
    def PreviewBoxWireframe(
        oBox_or_bbox: BaseBox | BBox3 | tuple[Vector3Like, Vector3Like],
        clrColor: ColorLike,
        *,
        oViewer: VedoViewer | None = None,
    ) -> list[int]:
        from ..base_shapes import BaseBox

        if isinstance(oBox_or_bbox, BaseBox):
            oBox = oBox_or_bbox
        elif isinstance(oBox_or_bbox, BBox3):
            oBox = BaseBox.from_bbox(oBox_or_bbox)
        else:
            oBox = BaseBox.from_bbox(cast(tuple[Vector3Like, Vector3Like], oBox_or_bbox))

        lines = [
            [oBox.vecGetSurfacePoint(-1, -1, 0), oBox.vecGetSurfacePoint(-1, 1, 0), oBox.vecGetSurfacePoint(1, 1, 0), oBox.vecGetSurfacePoint(1, -1, 0), oBox.vecGetSurfacePoint(-1, -1, 0)],
            [oBox.vecGetSurfacePoint(-1, -1, 1), oBox.vecGetSurfacePoint(-1, 1, 1), oBox.vecGetSurfacePoint(1, 1, 1), oBox.vecGetSurfacePoint(1, -1, 1), oBox.vecGetSurfacePoint(-1, -1, 1)],
            [oBox.vecGetSurfacePoint(-1, -1, 0), oBox.vecGetSurfacePoint(-1, -1, 1)],
            [oBox.vecGetSurfacePoint(-1, 1, 0), oBox.vecGetSurfacePoint(-1, 1, 1)],
            [oBox.vecGetSurfacePoint(1, 1, 0), oBox.vecGetSurfacePoint(1, 1, 1)],
            [oBox.vecGetSurfacePoint(1, -1, 0), oBox.vecGetSurfacePoint(1, -1, 1)],
        ]
        return [Sh.PreviewLine(line, clrColor, oViewer=oViewer) for line in lines]

    @staticmethod
    def PreviewRingWireframe(
        oRing: BaseRing,
        clrColor: ColorLike,
        nRadialSamples: int = 4,
        nLengthSamples: int = 10,
        *,
        oViewer: VedoViewer | None = None,
    ) -> list[int]:
        from ..base_shapes import BaseRing

        if not isinstance(oRing, BaseRing):
            raise TypeError("PreviewRingWireframe expects a BaseRing")
        groups: list[int] = []
        for nLengthSample in range(max(2, int(nLengthSamples))):
            fLengthRatio = float(nLengthSample) / float(max(1, nLengthSamples - 1))
            points = [oRing.vecGetSurfacePoint(fLengthRatio, float(i) / 99.0, 1.0) for i in range(100)]
            groups.append(Sh.PreviewLine(points, clrColor, oViewer=oViewer))
        for nRadialSample in range(max(2, int(nRadialSamples))):
            fPhiRatio = float(nRadialSample) / float(max(1, nRadialSamples - 1))
            points = [oRing.vecGetSurfacePoint(float(i) / 499.0, fPhiRatio, 1.0) for i in range(500)]
            groups.append(Sh.PreviewLine(points, clrColor, oViewer=oViewer))
        return groups

    @staticmethod
    def Preview(*args: object, oViewer: VedoViewer | None = None, **kwargs: object) -> object:
        if not args:
            raise TypeError("Preview expects at least one positional argument")

        obj = args[0]
        if isinstance(obj, Mesh):
            transparency, metallic, roughness = Sh._parse_preview_style(args, 2)
            return Sh.PreviewMesh(obj, cast(ColorLike, args[1]), transparency, metallic, roughness, oViewer=oViewer, **kwargs)
        if isinstance(obj, Voxels):
            transparency, metallic, roughness = Sh._parse_preview_style(args, 2)
            return Sh.PreviewVoxels(obj, cast(ColorLike, args[1]), transparency, metallic, roughness, oViewer=oViewer, **kwargs)
        if isinstance(obj, Lattice):
            transparency, metallic, roughness = Sh._parse_preview_style(args, 2)
            return Sh.PreviewLattice(obj, cast(ColorLike, args[1]), transparency, metallic, roughness, oViewer=oViewer, **kwargs)
        if isinstance(obj, LocalFrame):
            if len(args) >= 2 and isinstance(args[1], (int, float)):
                return Sh.PreviewFrame(obj, float(args[1]), oViewer=oViewer)
            raise TypeError("Preview(LocalFrame, ...) expects Preview(frame, size)")
        if isinstance(obj, BBox3):
            return Sh.PreviewBoxWireframe(obj, cast(ColorLike, args[1]), oViewer=oViewer)

        from ..base_shapes import BaseBox, BaseCylinder, BasePipe, BaseRing
        from ..frames import Frames

        if isinstance(obj, Frames):
            return Sh.PreviewFrames(obj, float(cast(float, args[1])), oViewer=oViewer)
        if isinstance(obj, BaseCylinder):
            return Sh.PreviewCylinderWireframe(obj, cast(ColorLike, args[1]), oViewer=oViewer)
        if isinstance(obj, BasePipe):
            return Sh.PreviewPipeWireframe(obj, cast(ColorLike, args[1]), oViewer=oViewer)
        if isinstance(obj, BaseBox):
            return Sh.PreviewBoxWireframe(obj, cast(ColorLike, args[1]), oViewer=oViewer)
        if isinstance(obj, BaseRing):
            return Sh.PreviewRingWireframe(obj, cast(ColorLike, args[1]), oViewer=oViewer)
        if Sh._is_point_like(obj) and len(args) >= 3 and isinstance(args[1], (int, float)):
            transparency, metallic, roughness = Sh._parse_preview_style(args, 3)
            return Sh.PreviewPoint(cast(Vector3Like, obj), float(cast(float, args[1])), cast(ColorLike, args[2]), transparency, metallic, roughness, oViewer=oViewer, **kwargs)
        if Sh._is_point_sequence(obj) and len(args) >= 2:
            return Sh.PreviewLine(cast(Sequence[Vector3Like], obj), cast(ColorLike, args[1]), oViewer=oViewer)

        raise TypeError(f"Unsupported Preview target: {type(obj)!r}")






from __future__ import annotations

import math

from shape_kernel._types import Vector3Like, as_np3, as_vec3

from picogk import IImplicit
from ..random_deformation_field import RandomDeformationField


class ImplicitRandomizedSchwarzPrimitive(IImplicit):
	def __init__(self, fUnitSize: float, fWallThickness: float, oField: RandomDeformationField) -> None:
		self.m_fFrequencyScale = (2.0 * math.pi) / float(fUnitSize)
		self.m_fWallThickness = float(fWallThickness)
		self.m_oRandomField = oField

	def fSignedDistance(self, vecPt: Vector3Like) -> float:
		vecNoise = as_np3(self.m_oRandomField.vecGetData(vecPt))
		vecNewPt = as_np3(vecPt) + vecNoise
		dX, dY, dZ = as_vec3(vecNewPt)
		dDist = (
			math.cos(self.m_fFrequencyScale * dX)
			+ math.cos(self.m_fFrequencyScale * dY)
			+ math.cos(self.m_fFrequencyScale * dZ)
		)
		return abs(dDist) - 0.5 * self.m_fWallThickness


__all__ = ["ImplicitRandomizedSchwarzPrimitive"]


from .beam_thickness import (
	BoundaryBeamThickness,
	CellBasedBeamThickness,
	ConstantBeamThickness,
	GlobalFuncBeamThickness,
	IBeamThickness,
)
from .cell_arrays import ConformalCellArray, ConformalShowcaseShapes, ICellArray, RegularCellArray, RegularUnitCell
from .lattice_types import BodyCentreLattice, ILatticeType, OctahedronLattice, RandomSplineLattice
from .unit_cells import CuboidCell, IUnitCell

__all__ = [
	"IBeamThickness",
	"BoundaryBeamThickness",
	"CellBasedBeamThickness",
	"ConstantBeamThickness",
	"GlobalFuncBeamThickness",
	"ICellArray",
	"ConformalCellArray",
	"ConformalShowcaseShapes",
	"RegularCellArray",
	"RegularUnitCell",
	"ILatticeType",
	"BodyCentreLattice",
	"OctahedronLattice",
	"RandomSplineLattice",
	"IUnitCell",
	"CuboidCell",
]


from __future__ import annotations

from collections.abc import Sequence

import numpy as np

from .._types import Vec3, Vector3Like, as_np3, as_vec3, clamp01


class ControlPointSurface:
    class EEnds:
        OPEN = "OPEN"
        CLOSED = "CLOSED"

    def __init__(
        self,
        aControlGrid: Sequence[Sequence[Vector3Like]],
        nDegreeU: int = 2,
        nDegreeV: int = 2,
        eEndsU: str = EEnds.OPEN,
        eEndsV: str = EEnds.OPEN,
    ) -> None:
        self.m_aControlGrid = [[as_np3(p) for p in row] for row in aControlGrid]
        if not self.m_aControlGrid or not self.m_aControlGrid[0]:
            raise ValueError("aControlGrid cannot be empty")

        self.m_nDegreeU = int(nDegreeU)
        self.m_nDegreeV = int(nDegreeV)
        self.m_eEndsU = eEndsU
        self.m_eEndsV = eEndsV
        self.m_fError = 1e-7

        if self.m_eEndsU == self.EEnds.CLOSED and len(self.m_aControlGrid) > 1:
            if float(np.linalg.norm(self.m_aControlGrid[0][0] - self.m_aControlGrid[-1][0])) < self.m_fError:
                self.m_aControlGrid.pop()

        if self.m_eEndsV == self.EEnds.CLOSED and len(self.m_aControlGrid[0]) > 1:
            if float(np.linalg.norm(self.m_aControlGrid[0][0] - self.m_aControlGrid[0][-1])) < self.m_fError:
                self.m_aControlGrid = [row[:-1] for row in self.m_aControlGrid]

        self.m_aKnotU = self.aGetUKnotVector(len(self.m_aControlGrid), self.m_nDegreeU, self.m_eEndsU)
        self.m_aKnotV = self.aGetVKnotVector(len(self.m_aControlGrid[0]), self.m_nDegreeV, self.m_eEndsV)

    def aGetUKnotVector(self, nNumberOfControlPoints: int, nDegree: int, eEnds: str) -> list[float]:
        if eEnds == self.EEnds.CLOSED:
            before = len(self.m_aControlGrid)
            for i in range(before - 1):
                self.m_aControlGrid.append([p.copy() for p in self.m_aControlGrid[i]])
            nNumberOfControlPoints = len(self.m_aControlGrid)

        n_knots = int(nNumberOfControlPoints) + int(nDegree) + 1
        n_valid_range = n_knots - int(nDegree) - (int(nDegree) + 1)
        d_r = 1.0 / float(n_valid_range)
        out: list[float] = []
        for i in range(n_knots):
            value = -(d_r * nDegree) + d_r * i
            if eEnds == self.EEnds.OPEN:
                value = clamp01(value)
            out.append(float(value))
        return out

    def aGetVKnotVector(self, nNumberOfControlPoints: int, nDegree: int, eEnds: str) -> list[float]:
        if eEnds == self.EEnds.CLOSED:
            before = len(self.m_aControlGrid[0])
            for row_idx in range(len(self.m_aControlGrid)):
                row = self.m_aControlGrid[row_idx]
                for i in range(before - 1):
                    row.append(row[i].copy())
            nNumberOfControlPoints = len(self.m_aControlGrid[0])

        n_knots = int(nNumberOfControlPoints) + int(nDegree) + 1
        n_valid_range = n_knots - int(nDegree) - (int(nDegree) + 1)
        d_r = 1.0 / float(n_valid_range)
        out: list[float] = []
        for i in range(n_knots):
            value = -(d_r * nDegree) + d_r * i
            if eEnds == self.EEnds.OPEN:
                value = clamp01(value)
            out.append(float(value))
        return out

    def aGetGrid(self, nUSamples: int = 500, nVSamples: int = 500) -> list[list[Vec3]]:
        grid: list[list[Vec3]] = []
        for i in range(int(nUSamples)):
            u_ratio = float(i) / float(max(1, nUSamples - 1))
            row: list[Vec3] = []
            for j in range(int(nVSamples)):
                v_ratio = float(j) / float(max(1, nVSamples - 1))
                row.append(self.vecGetPointAt(u_ratio, v_ratio))
            grid.append(row)
        return grid

    def vecGetPointAt(self, fURatio: float, fVRatio: float) -> Vec3:
        point = np.zeros(3, dtype=np.float64)
        for u in range(len(self.m_aControlGrid)):
            u_base = self.fBaseFunc(self.m_aKnotU, fURatio, u, self.m_nDegreeU)
            for v in range(len(self.m_aControlGrid[u])):
                v_base = self.fBaseFunc(self.m_aKnotV, fVRatio, v, self.m_nDegreeV)
                point += u_base * v_base * self.m_aControlGrid[u][v]
        return as_vec3(point)

    def fBaseFunc(self, aKnot: Sequence[float], fLengthRatio: float, nControlPoint: int, nDegree: int) -> float:
        if nDegree == 0:
            left = aKnot[nControlPoint]
            right = aKnot[nControlPoint + 1]
            if (left <= fLengthRatio < right) or (
                abs(fLengthRatio - right) < self.m_fError and abs(fLengthRatio - aKnot[-1]) < self.m_fError
            ):
                return 1.0
            return 0.0

        value = 0.0
        left_den = aKnot[nControlPoint + nDegree] - aKnot[nControlPoint]
        if abs(left_den) > self.m_fError:
            value += (
                (fLengthRatio - aKnot[nControlPoint])
                / left_den
                * self.fBaseFunc(aKnot, fLengthRatio, nControlPoint, nDegree - 1)
            )

        right_den = aKnot[nControlPoint + nDegree + 1] - aKnot[nControlPoint + 1]
        if abs(right_den) > self.m_fError:
            value += (
                (aKnot[nControlPoint + nDegree + 1] - fLengthRatio)
                / right_den
                * self.fBaseFunc(aKnot, fLengthRatio, nControlPoint + 1, nDegree - 1)
            )

        return float(value)

    def aGetControlGrid(self) -> list[list[Vec3]]:
        return [[as_vec3(p) for p in row] for row in self.m_aControlGrid]

    def vecGetControlPoint(self, u: int, v: int) -> Vec3:
        return as_vec3(self.m_aControlGrid[int(u)][int(v)])

    def UpdateControlPoint(self, vecPt: Vector3Like, u: int, v: int) -> None:
        self.m_aControlGrid[int(u)][int(v)] = as_np3(vecPt)




from __future__ import annotations

import math
from typing import TYPE_CHECKING

import numpy as np

from .._types import Vec3, Vector3Like, as_np3, as_vec3, normalized

if TYPE_CHECKING:
	from ..frames.local_frames import LocalFrame


class VecOperations:
	@staticmethod
	def R(vecPt: Vector3Like) -> float:
		return VecOperations.fGetRadius(vecPt)

	@staticmethod
	def Normalize(vecA: Vector3Like) -> Vec3:
		return as_vec3(normalized(vecA))

	@staticmethod
	def ConvertTo3D(vecFlat: tuple[float, float] | list[float], fZ: float = 0.0) -> Vec3:
		if len(vecFlat) < 2:
			raise ValueError("vecFlat must contain at least 2 coordinates")
		return as_vec3((float(vecFlat[0]), float(vecFlat[1]), float(fZ)))

	@staticmethod
	def ConvertTo2D(vecA: Vector3Like) -> tuple[float, float]:
		vec = as_np3(vecA)
		return (float(vec[0]), float(vec[1]))

	@staticmethod
	def vecGetOrthogonalDir(vecDir: Vector3Like) -> np.ndarray:
		direction = normalized(vecDir)
		non_parallel = np.array([1.0, 0.0, 0.0], dtype=np.float64)
		if abs(float(np.dot(direction, non_parallel))) > 0.95:
			non_parallel = np.array([0.0, 1.0, 0.0], dtype=np.float64)
		return normalized(np.cross(direction, non_parallel))

	@staticmethod
	def vecFlipForAlignment(vecDir: Vector3Like, vecTargetDir: Vector3Like) -> np.ndarray:
		d = as_np3(vecDir)
		t = as_np3(vecTargetDir)
		return d if float(np.dot(t, d)) >= float(np.dot(t, -d)) else -d

	@staticmethod
	def fGetAngleBetween(vecA: Vector3Like, vecB: Vector3Like) -> float:
		a = normalized(vecA)
		b = normalized(vecB)
		dot = float(np.clip(np.dot(a, b), -1.0, 1.0))
		theta = float(math.acos(dot))
		if math.isnan(theta) and abs(dot) == 1.0:
			return math.pi
		return theta

	@staticmethod
	def vecRotateAroundAxis(
		vecPt: Vector3Like,
		dPhi: float,
		vecAxis: Vector3Like,
		vecAxisOrigin: Vector3Like | None = None,
	) -> Vec3:
		point = as_np3(vecPt)
		axis = normalized(vecAxis)
		origin = np.zeros(3, dtype=np.float64) if vecAxisOrigin is None else as_np3(vecAxisOrigin)
		rel = point - origin
		cos_phi = math.cos(dPhi)
		sin_phi = math.sin(dPhi)
		rot = rel * cos_phi + np.cross(axis, rel) * sin_phi + axis * float(np.dot(axis, rel)) * (1.0 - cos_phi)
		return as_vec3(rot + origin)

	@staticmethod
	def vecTranslatePointOntoFrame(oFrame: LocalFrame, vecPt: Vector3Like) -> Vec3:
		p = as_np3(vecPt)
		origin = as_np3(oFrame.vecGetPosition())
		x = as_np3(oFrame.vecGetLocalX())
		y = as_np3(oFrame.vecGetLocalY())
		z = as_np3(oFrame.vecGetLocalZ())
		return as_vec3(origin + p[0] * x + p[1] * y + p[2] * z)

	@staticmethod
	def vecGetPlanarDir(vecPt: Vector3Like) -> np.ndarray:
		pt = as_np3(vecPt)
		planar = np.array([pt[0], pt[1], 0.0], dtype=np.float64)
		length = float(np.linalg.norm(planar))
		if length <= 1e-12:
			return np.array([1.0, 0.0, 0.0], dtype=np.float64)
		return planar / length

	@staticmethod
	def vecSetZ(vecPt: Vector3Like, zValue: float) -> Vec3:
		pt = as_np3(vecPt).copy()
		pt[2] = float(zValue)
		return as_vec3(pt)

	@staticmethod
	def vecSetRadius(vecPt: Vector3Like, radius: float) -> Vec3:
		pt = as_np3(vecPt).copy()
		planar = np.array([pt[0], pt[1]], dtype=np.float64)
		length = float(np.linalg.norm(planar))
		target = float(radius)
		if length <= 1e-12:
			pt[0] = target
			pt[1] = 0.0
			return as_vec3(pt)
		scale = target / length
		pt[0] *= scale
		pt[1] *= scale
		return as_vec3(pt)

	@staticmethod
	def vecGetCylPoint(fRadius: float, fPhi: float, fZ: float) -> Vec3:
		r = float(fRadius)
		return as_vec3(np.array([r * math.cos(float(fPhi)), r * math.sin(float(fPhi)), float(fZ)], dtype=np.float64))

	@staticmethod
	def vecGetSphPoint(fRadius: float, fPhi: float, fTheta: float) -> Vec3:
		r, phi, theta = float(fRadius), float(fPhi), float(fTheta)
		return as_vec3(np.array([r * math.cos(phi) * math.cos(theta), r * math.sin(phi) * math.cos(theta), r * math.sin(theta)], dtype=np.float64))

	@staticmethod
	def fGetRadius(vecPt: Vector3Like) -> float:
		pt = as_np3(vecPt)
		return float(math.sqrt(float(pt[0]) ** 2 + float(pt[1]) ** 2))

	@staticmethod
	def fGetPhi(vecPt: Vector3Like) -> float:
		pt = as_np3(vecPt)
		return float(math.atan2(float(pt[1]), float(pt[0])))

	@staticmethod
	def fGetTheta(vecPt: Vector3Like) -> float:
		r = VecOperations.fGetRadius(vecPt)
		return float(math.atan2(float(as_np3(vecPt)[2]), r))

	@staticmethod
	def vecSetPhi(vecPt: Vector3Like, fNewPhi: float) -> Vec3:
		r = VecOperations.fGetRadius(vecPt)
		z = float(as_np3(vecPt)[2])
		return VecOperations.vecGetCylPoint(r, float(fNewPhi), z)

	@staticmethod
	def vecUpdateRadius(vecPt: Vector3Like, dRadius: float) -> Vec3:
		new_r = VecOperations.fGetRadius(vecPt) + float(dRadius)
		phi = VecOperations.fGetPhi(vecPt)
		z = float(as_np3(vecPt)[2])
		return VecOperations.vecGetCylPoint(new_r, phi, z)

	@staticmethod
	def vecUpdatePhi(vecPt: Vector3Like, dPhi: float) -> Vec3:
		r = VecOperations.fGetRadius(vecPt)
		new_phi = VecOperations.fGetPhi(vecPt) + float(dPhi)
		z = float(as_np3(vecPt)[2])
		return VecOperations.vecGetCylPoint(r, new_phi, z)

	@staticmethod
	def vecUpdateZ(vecPt: Vector3Like, dZ: float) -> Vec3:
		return VecOperations.vecSetZ(vecPt, float(as_np3(vecPt)[2]) + float(dZ))

	@staticmethod
	def vecRotateAroundZ(vecPt: Vector3Like, dPhi: float, vecAxisOrigin: Vector3Like | None = None) -> Vec3:
		origin = np.zeros(3, dtype=np.float64) if vecAxisOrigin is None else as_np3(vecAxisOrigin)
		diff = as_np3(vecPt) - origin
		phi = float(math.atan2(float(diff[1]), float(diff[0])))
		r = float(math.sqrt(float(diff[0]) ** 2 + float(diff[1]) ** 2))
		new_diff = np.array([r * math.cos(phi + float(dPhi)), r * math.sin(phi + float(dPhi)), float(diff[2])], dtype=np.float64)
		return as_vec3(new_diff + origin)

	@staticmethod
	def vecExpressPointInFrame(oFrame: LocalFrame, vecPt: Vector3Like) -> Vec3:
		pt = as_np3(vecPt) - as_np3(oFrame.vecGetPosition())
		x = float(np.dot(pt, as_np3(oFrame.vecGetLocalX())))
		y = float(np.dot(pt, as_np3(oFrame.vecGetLocalY())))
		z = float(np.dot(pt, as_np3(oFrame.vecGetLocalZ())))
		return (x, y, z)

	@staticmethod
	def vecTranslateDirectionOntoFrame(oFrame: LocalFrame, vecDir: Vector3Like) -> Vec3:
		origin = np.zeros(3, dtype=np.float64)
		pt1 = as_np3(VecOperations.vecTranslatePointOntoFrame(oFrame, origin))
		pt2 = as_np3(VecOperations.vecTranslatePointOntoFrame(oFrame, as_np3(vecDir)))
		return as_vec3(pt2 - pt1)

	@staticmethod
	def vecGetDirectionToAxis(oFrame: LocalFrame, vecPt: Vector3Like) -> Vec3:
		pt = as_np3(vecPt) - as_np3(oFrame.vecGetPosition())
		lx = as_np3(oFrame.vecGetLocalX())
		ly = as_np3(oFrame.vecGetLocalY())
		d = float(np.dot(pt, lx)) * lx + float(np.dot(pt, ly)) * ly
		mag = float(np.linalg.norm(d))
		return as_vec3(d / mag) if mag > 1e-12 else as_vec3(lx)

	@staticmethod
	def fGetRadiusToAxis(oFrame: LocalFrame, vecPt: Vector3Like) -> float:
		pt = as_np3(vecPt) - as_np3(oFrame.vecGetPosition())
		lx = as_np3(oFrame.vecGetLocalX())
		ly = as_np3(oFrame.vecGetLocalY())
		return float(math.sqrt(float(np.dot(pt, lx)) ** 2 + float(np.dot(pt, ly)) ** 2))

	@staticmethod
	def fGetPhiToAxis(oFrame: LocalFrame, vecPt: Vector3Like) -> float:
		pt = as_np3(vecPt) - as_np3(oFrame.vecGetPosition())
		lx = as_np3(oFrame.vecGetLocalX())
		ly = as_np3(oFrame.vecGetLocalY())
		return float(math.atan2(float(np.dot(pt, ly)), float(np.dot(pt, lx))))

	@staticmethod
	def vecLinearInterpolation(vecPt1: Vector3Like, vecPt2: Vector3Like, fRatio: float) -> Vec3:
		p1 = as_np3(vecPt1)
		p2 = as_np3(vecPt2)
		return as_vec3(p1 + float(fRatio) * (p2 - p1))

	@staticmethod
	def vecCylindricalInterpolation(
		vecPt1: Vector3Like,
		vecPt2: Vector3Like,
		fRatio: float,
		vecAxisOrigin: Vector3Like | None = None,
	) -> Vec3:
		origin = np.zeros(3, dtype=np.float64) if vecAxisOrigin is None else as_np3(VecOperations.vecSetZ(vecAxisOrigin, 0.0))
		p1 = as_np3(vecPt1)
		p2 = as_np3(vecPt2)
		d_angle = VecOperations.fGetAngleBetween(p1 - origin, p2 - origin)
		side1 = normalized(p1 - origin)
		r_pos = as_np3(VecOperations.vecRotateAroundZ(as_vec3(p1), d_angle))
		r_neg = as_np3(VecOperations.vecRotateAroundZ(as_vec3(p1), -d_angle))
		sense = 1 if float(np.linalg.norm(p2 - r_neg)) >= float(np.linalg.norm(p2 - r_pos)) else -1
		r1 = VecOperations.fGetRadius(as_vec3(p1 - origin))
		r2 = VecOperations.fGetRadius(as_vec3(p2 - origin))
		ir = r1 + float(fRatio) * (r2 - r1)
		iz = float(p1[2]) + float(fRatio) * (float(p2[2]) - float(p1[2]))
		inter = as_np3(VecOperations.vecRotateAroundZ(as_vec3(ir * side1), sense * float(fRatio) * d_angle))
		return as_vec3(as_np3(VecOperations.vecSetZ(as_vec3(inter), iz)) + origin)

	@staticmethod
	def vecSphericalInterpolation(
		vecPt1: Vector3Like,
		vecPt2: Vector3Like,
		fRatio: float,
		vecAxisOrigin: Vector3Like | None = None,
	) -> Vec3:
		origin = np.zeros(3, dtype=np.float64) if vecAxisOrigin is None else as_np3(vecAxisOrigin)
		p1 = as_np3(vecPt1)
		p2 = as_np3(vecPt2)
		d_angle = VecOperations.fGetAngleBetween(p1 - origin, p2 - origin)
		side1 = normalized(p1 - origin)
		normal = np.cross(side1, normalized(p2 - origin))
		r_pos = as_np3(VecOperations.vecRotateAroundAxis(as_vec3(p1), d_angle, as_vec3(normal)))
		r_neg = as_np3(VecOperations.vecRotateAroundAxis(as_vec3(p1), -d_angle, as_vec3(normal)))
		sense = 1 if float(np.linalg.norm(p2 - r_neg)) >= float(np.linalg.norm(p2 - r_pos)) else -1
		r1 = float(np.linalg.norm(p1 - origin))
		r2 = float(np.linalg.norm(p2 - origin))
		ir = r1 + float(fRatio) * (r2 - r1)
		inter = as_np3(VecOperations.vecRotateAroundAxis(as_vec3(ir * side1), sense * float(fRatio) * d_angle, as_vec3(normal)))
		return as_vec3(inter + origin)

	@staticmethod
	def fGetSignedAngleBetween(vecA: Vector3Like, vecB: Vector3Like, vecRefNormal: Vector3Like) -> float:
		a, b, ref = as_np3(vecA), as_np3(vecB), as_np3(vecRefNormal)
		if float(np.dot(a, a)) < 1e-20 or float(np.dot(b, b)) < 1e-20 or float(np.dot(ref, ref)) < 1e-20:
			raise ValueError("Vector3 with zero length.")
		a, b, ref = normalized(a), normalized(b), normalized(ref)
		normal = as_np3(VecOperations.vecFlipForAlignment(np.cross(a, b), ref))
		theta = abs(VecOperations.fGetAngleBetween(a, b))
		r_pos = as_np3(VecOperations.vecRotateAroundAxis(as_vec3(b), theta, as_vec3(normal)))
		r_neg = as_np3(VecOperations.vecRotateAroundAxis(as_vec3(b), -theta, as_vec3(normal)))
		return theta if float(np.dot(a, r_neg)) <= float(np.dot(a, r_pos)) else -theta

	@staticmethod
	def bCheckAlignment(vecDir: Vector3Like, vecTargetDir: Vector3Like) -> bool:
		d = as_np3(vecDir)
		t = as_np3(vecTargetDir)
		return float(np.dot(t, d)) >= float(np.dot(t, -d))


__all__ = ["VecOperations"]



from __future__ import annotations

import os
import platform
from pathlib import Path
from typing import Iterable

DEFAULT_RUNTIME_NAME = "picogk.1.7"
STRING_LENGTH = 255


def _platform_suffixes() -> list[str]:
    system = platform.system().lower()
    if system == "windows":
        return [".dll"]
    if system == "darwin":
        return [".dylib", ".so"]
    return [".so", ".dylib"]


def runtime_candidates(runtime_name: str = DEFAULT_RUNTIME_NAME) -> list[str]:
    suffixes = _platform_suffixes()
    candidates: list[str] = [runtime_name]
    for suffix in suffixes:
        candidates.append(f"{runtime_name}{suffix}")
        if not runtime_name.startswith("lib"):
            candidates.append(f"lib{runtime_name}{suffix}")
    return candidates


def _platform_native_subdirs() -> list[str]:
    system = platform.system().lower()
    machine = platform.machine().lower()

    if system == "windows":
        return ["win-x64"]

    if system == "darwin":
        if "arm" in machine or "aarch64" in machine:
            return ["osx-arm64"]
        return ["osx-x64", "osx-arm64"]

    if "arm" in machine or "aarch64" in machine:
        return ["linux-arm64", "linux-x64"]
    return ["linux-x64", "linux-arm64"]


def _bundled_native_dirs() -> Iterable[Path]:
    base = Path(__file__).resolve().parent / "native"
    if not base.exists():
        return

    yielded = False
    for subdir in _platform_native_subdirs():
        p = base / subdir
        if p.exists():
            yielded = True
            yield p

    if not yielded:
        for p in base.iterdir():
            if p.is_dir():
                yield p


def candidate_library_paths(runtime_name: str = DEFAULT_RUNTIME_NAME) -> list[str]:
    env = os.getenv("PICOGK_RUNTIME_PATH")
    if env:
        return [env]

    candidates = runtime_candidates(runtime_name)
    paths = list(candidates)

    for native_dir in _bundled_native_dirs():
        for child in native_dir.rglob("*"):
            if not child.is_file():
                continue
            if child.name in candidates:
                paths.append(str(child))

    return paths




from typing import Optional
import random

def _hex_to_rgba(hex_value: str) -> tuple[float, float, float, float]:
    value = hex_value.strip().lstrip("#")
    if len(value) != 6:
        raise ValueError("Expected RGB hex color")
    r = int(value[0:2], 16) / 255.0
    g = int(value[2:4], 16) / 255.0
    b = int(value[4:6], 16) / 255.0
    return (r, g, b, 1.0)

class Cp:
    clrBlack = _hex_to_rgba("#000000")
    clrRacingGreen = _hex_to_rgba("#065c35")
    clrGreen = _hex_to_rgba("#00b800")
    clrBillie = _hex_to_rgba("#02f70b")
    clrLemongrass = _hex_to_rgba("#b8e031")
    clrYellow = _hex_to_rgba("#fcd808")
    clrWarning = _hex_to_rgba("#fc6608")
    clrRed = _hex_to_rgba("#ff0000")
    clrRuby = _hex_to_rgba("#b0002c")
    clrOrchid = _hex_to_rgba("#c72483")
    clrPitaya = _hex_to_rgba("#fa2a88")
    clrBubblegum = _hex_to_rgba("#ff66ce")
    clrLavender = _hex_to_rgba("#c966ff")
    clrGray = _hex_to_rgba("#bdbdbd")
    clrRock = _hex_to_rgba("#6b7178")
    clrCrystal = _hex_to_rgba("#0cc1f7")
    clrFrozen = _hex_to_rgba("#6de2fc")
    clrBlueberry = _hex_to_rgba("#4f0dbf")
    clrBlue = _hex_to_rgba("#4287f5")
    clrToothpaste = _hex_to_rgba("#25e6c9")

    def clrRandom(self, j: Optional[int] = None) -> tuple[float, float, float, float]:
        rng = random.Random(j)
        return (float(rng.random()), float(rng.random()), float(rng.random()), 1.0)



from __future__ import annotations

from enum import IntEnum
from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from .vedo_viewer import VedoViewer


class IViewerAction(Protocol):
    def Do(self, viewer: "VedoViewer") -> None:
        ...


class IKeyHandler(Protocol):
    def bHandleEvent(
        self,
        viewer: "VedoViewer",
        eKey: str,
        bPressed: bool,
        bShift: bool,
        bCtrl: bool,
        bAlt: bool,
        bCmd: bool,
    ) -> bool:
        ...


class KeyAction:
    def __init__(
        self,
        xAction: IViewerAction,
        eKey: str,
        bPressed: bool = False,
        bShift: bool = False,
        bCtrl: bool = False,
        bAlt: bool = False,
        bCmd: bool = False,
    ) -> None:
        self._action = xAction
        self._key = str(eKey).lower()
        self._pressed = bool(bPressed)
        self._shift = bool(bShift)
        self._ctrl = bool(bCtrl)
        self._alt = bool(bAlt)
        self._cmd = bool(bCmd)

    def bKeyEquals(
        self,
        eKey: str,
        bPressed: bool,
        bShift: bool,
        bCtrl: bool,
        bAlt: bool,
        bCmd: bool,
    ) -> bool:
        return (
            self._key == str(eKey).lower()
            and self._pressed == bool(bPressed)
            and self._shift == bool(bShift)
            and self._ctrl == bool(bCtrl)
            and self._alt == bool(bAlt)
            and self._cmd == bool(bCmd)
        )

    def Do(self, viewer: "VedoViewer") -> None:
        self._action.Do(viewer)


class KeyHandler:
    def __init__(self) -> None:
        self._actions: list[KeyAction] = []

    def AddAction(self, action: KeyAction) -> None:
        self._actions.insert(0, action)

    def bHandleEvent(
        self,
        viewer: "VedoViewer",
        eKey: str,
        bPressed: bool,
        bShift: bool,
        bCtrl: bool,
        bAlt: bool,
        bCmd: bool,
    ) -> bool:
        for action in self._actions:
            if action.bKeyEquals(eKey, bPressed, bShift, bCtrl, bAlt, bCmd):
                action.Do(viewer)
                return True
        return False


class RotateToNextRoundAngleAction:
    class EDir(IntEnum):
        Dir_Down = 0
        Dir_Up = 1
        Dir_Left = 2
        Dir_Right = 3

    def __init__(self, eDir: "RotateToNextRoundAngleAction.EDir") -> None:
        self._dir = eDir

    def Do(self, viewer: "VedoViewer") -> None:
        step = 15.0
        if self._dir == self.EDir.Dir_Left:
            viewer.AdjustViewAngles(-step, 0.0)
        elif self._dir == self.EDir.Dir_Right:
            viewer.AdjustViewAngles(step, 0.0)
        elif self._dir == self.EDir.Dir_Up:
            viewer.AdjustViewAngles(0.0, step)
        else:
            viewer.AdjustViewAngles(0.0, -step)






from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from ..unit_cells.i_unit_cell import IUnitCell


class ICellArray(ABC):
	@abstractmethod
	def aGetUnitCells(self) -> list[IUnitCell]:
		raise NotImplementedError


__all__ = ["ICellArray"]


from __future__ import annotations

import time
from dataclasses import dataclass
from pathlib import Path

@dataclass
class _TimeLapseState:
    interval_ms: float
    path: Path
    file_name: str
    current_frame: int
    paused: bool
    next_due_ms: float

    @classmethod
    def create(
        cls,
        fIntervalInMilliseconds: float,
        strPath: str,
        strFileName: str,
        nStartFrame: int,
        bPaused: bool,
    ) -> "_TimeLapseState":
        now = time.perf_counter() * 1000.0
        path = Path(strPath)
        path.mkdir(parents=True, exist_ok=True)
        return cls(
            interval_ms=float(fIntervalInMilliseconds),
            path=path,
            file_name=strFileName,
            current_frame=int(nStartFrame),
            paused=bool(bPaused),
            next_due_ms=now + float(fIntervalInMilliseconds),
        )

    def Pause(self) -> None:
        self.paused = True

    def Resume(self) -> None:
        self.paused = False
        self.next_due_ms = (time.perf_counter() * 1000.0) + self.interval_ms

    def bDue(self) -> tuple[bool, str]:
        if self.paused:
            return False, ""
        now = time.perf_counter() * 1000.0
        if now < self.next_due_ms:
            return False, ""
        frame_path = self.path / f"{self.file_name}{self.current_frame:05d}.png"
        self.current_frame += 1
        self.next_due_ms = now + self.interval_ms
        return True, str(frame_path)






from __future__ import annotations

from dataclasses import dataclass
import math

from _common.types import Vector2Like, Vector3Like

from .._types import Vec3, _BBox3C, as_vec3, vec3_tuple


@dataclass
class BBox2:
    vecMin: tuple[float, float] = (float("inf"), float("inf"))
    vecMax: tuple[float, float] = (float("-inf"), float("-inf"))

    def __post_init__(self) -> None:
        self.vecMin = (float(self.vecMin[0]), float(self.vecMin[1]))
        self.vecMax = (float(self.vecMax[0]), float(self.vecMax[1]))
        if not self.bIsEmpty():
            if self.vecMin[0] > self.vecMax[0] or self.vecMin[1] > self.vecMax[1]:
                raise ValueError("BBox2 requires vecMin <= vecMax for non-empty boxes")

    def bIsEmpty(self) -> bool:
        return (
            math.isinf(self.vecMin[0])
            and self.vecMin[0] > 0.0
            and math.isinf(self.vecMin[1])
            and self.vecMin[1] > 0.0
            and math.isinf(self.vecMax[0])
            and self.vecMax[0] < 0.0
            and math.isinf(self.vecMax[1])
            and self.vecMax[1] < 0.0
        )

    def bContains(self, point: Vector2Like) -> bool:
        if self.bIsEmpty():
            return False
        x, y = float(point[0]), float(point[1])
        return (
            self.vecMin[0] <= x <= self.vecMax[0]
            and self.vecMin[1] <= y <= self.vecMax[1]
        )

    def Include(self, point: Vector2Like | "BBox2") -> None:
        if isinstance(point, BBox2):
            if point.bIsEmpty():
                return
            self.Include(point.vecMin)
            self.Include(point.vecMax)
            return
        x, y = float(point[0]), float(point[1])
        self.vecMin = (min(self.vecMin[0], x), min(self.vecMin[1], y))
        self.vecMax = (max(self.vecMax[0], x), max(self.vecMax[1], y))

    def Grow(self, fGrowBy: float) -> None:
        grow = float(fGrowBy)
        if self.bIsEmpty():
            self.vecMin = (-grow, -grow)
            self.vecMax = (grow, grow)
            return
        self.vecMin = (self.vecMin[0] - grow, self.vecMin[1] - grow)
        self.vecMax = (self.vecMax[0] + grow, self.vecMax[1] + grow)

    def vecSize(self) -> tuple[float, float]:
        return (self.vecMax[0] - self.vecMin[0], self.vecMax[1] - self.vecMin[1])

    def vecCenter(self) -> tuple[float, float]:
        size = self.vecSize()
        return (self.vecMin[0] + 0.5 * size[0], self.vecMin[1] + 0.5 * size[1])

    def __str__(self) -> str:
        return f"<Min: {self.vecMin} | Max: {self.vecMax}>"


class BBox3(_BBox3C):
    def __init__(
        self,
        min_xyz: Vector3Like | Vec3 | None = None,
        max_xyz: Vector3Like | Vec3 | None = None,
    ) -> None:
        if min_xyz is None and max_xyz is None:
            super().__init__(
                Vec3(float("inf"), float("inf"), float("inf")),
                Vec3(float("-inf"), float("-inf"), float("-inf")),
            )
            return

        if min_xyz is None or max_xyz is None:
            raise ValueError("BBox3 requires both min_xyz and max_xyz, or neither")

        vmin = as_vec3(min_xyz)
        vmax = as_vec3(max_xyz)
        if vmin.x > vmax.x or vmin.y > vmax.y or vmin.z > vmax.z:
            raise ValueError("BBox3 requires min_xyz <= max_xyz")
        super().__init__(vmin, vmax)

    @property
    def vecMin(self) -> Vec3:
        return self.min

    @vecMin.setter
    def vecMin(self, value: Vector3Like | Vec3) -> None:
        self.min = as_vec3(value)

    @property
    def vecMax(self) -> Vec3:
        return self.max

    @vecMax.setter
    def vecMax(self, value: Vector3Like | Vec3) -> None:
        self.max = as_vec3(value)

    def bIsEmpty(self) -> bool:
        return (
            math.isinf(self.min.x)
            and self.min.x > 0.0
            and math.isinf(self.min.y)
            and self.min.y > 0.0
            and math.isinf(self.min.z)
            and self.min.z > 0.0
            and math.isinf(self.max.x)
            and self.max.x < 0.0
            and math.isinf(self.max.y)
            and self.max.y < 0.0
            and math.isinf(self.max.z)
            and self.max.z < 0.0
        )

    def bContains(self, point: Vector3Like | Vec3) -> bool:
        if self.bIsEmpty():
            return False
        v = as_vec3(point)
        return (
            self.min.x <= v.x <= self.max.x
            and self.min.y <= v.y <= self.max.y
            and self.min.z <= v.z <= self.max.z
        )

    def Include(self, item: "BBox3 | BBox2 | Vector3Like | Vec3", fZ: float = 0.0) -> None:
        if isinstance(item, BBox3):
            if item.bIsEmpty():
                return
            self.Include(item.min)
            self.Include(item.max)
            return

        if isinstance(item, BBox2):
            if item.bIsEmpty():
                return
            self.Include((item.vecMin[0], item.vecMin[1], float(fZ)))
            self.Include((item.vecMax[0], item.vecMax[1], float(fZ)))
            return

        v = as_vec3(item)
        self.min = Vec3(min(self.min.x, v.x), min(self.min.y, v.y), min(self.min.z, v.z))
        self.max = Vec3(max(self.max.x, v.x), max(self.max.y, v.y), max(self.max.z, v.z))

    def Grow(self, fGrowBy: float) -> None:
        grow = float(fGrowBy)
        if self.bIsEmpty():
            self.min = Vec3(-grow, -grow, -grow)
            self.max = Vec3(grow, grow, grow)
            return
        self.min = Vec3(self.min.x - grow, self.min.y - grow, self.min.z - grow)
        self.max = Vec3(self.max.x + grow, self.max.y + grow, self.max.z + grow)

    def vecSize(self) -> tuple[float, float, float]:
        return (
            float(self.max.x - self.min.x),
            float(self.max.y - self.min.y),
            float(self.max.z - self.min.z),
        )

    def vecCenter(self) -> tuple[float, float, float]:
        sx, sy, sz = self.vecSize()
        return (
            float(self.min.x + 0.5 * sx),
            float(self.min.y + 0.5 * sy),
            float(self.min.z + 0.5 * sz),
        )

    def oAsBoundingBox2(self) -> BBox2:
        return BBox2(
            (float(self.min.x), float(self.min.y)),
            (float(self.max.x), float(self.max.y)),
        )

    def __str__(self) -> str:
        return f"<Min: {vec3_tuple(self.min)} | Max: {vec3_tuple(self.max)}>"




from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from picogk import Lattice

if TYPE_CHECKING:
	from ..beam_thickness.i_beam_thickness import IBeamThickness
	from ..unit_cells.i_unit_cell import IUnitCell


class ILatticeType(ABC):
	@abstractmethod
	def AddCell(self, oLattice: Lattice, xCell: IUnitCell, xBeamThickness: IBeamThickness, nSubSamples: int) -> None:
		raise NotImplementedError


__all__ = ["ILatticeType"]


from __future__ import annotations

from abc import ABC, abstractmethod

from picogk import BBox3
from shape_kernel._types import Vec3


class IUnitCell(ABC):
	@abstractmethod
	def aGetCornerPoints(self) -> list[Vec3]:
		raise NotImplementedError

	@abstractmethod
	def vecGetCellCentre(self) -> tuple[float, float, float]:
		raise NotImplementedError

	@abstractmethod
	def oGetCellBounding(self) -> BBox3:
		raise NotImplementedError

	@abstractmethod
	def PreviewUnitCell(self) -> None:
		raise NotImplementedError


__all__ = ["IUnitCell"]


from __future__ import annotations

from enum import Enum
from typing import Callable

import numpy as np

from picogk import Mesh
from shape_kernel import LocalFrame, MeshUtility, Sh, VecOperations
from shape_kernel._types import Vec3, as_np3, as_vec3, normalized

from .icosahedral_face import IcosehedralFace


class QuasiTile:
    class EPreviewFace(Enum):
        NONE = "NONE"
        AXIS = "AXIS"
        CONNECTOR = "CONNECTOR"

    def __init__(self) -> None:
        self.m_clr = (1.0, 1.0, 1.0, 1.0)
        self.m_aFaces: list[IcosehedralFace] = []
        self.m_vecRoundedCentre: Vec3 | None = None

    def aGetFaces(self) -> list[IcosehedralFace]:
        return self.m_aFaces

    def ApplyTrafo(self, oTrafoFunc: Callable[[Vec3], Vec3]) -> None:
        for s_face in self.m_aFaces:
            s_face.vecPt1 = oTrafoFunc(s_face.vecPt1)
            s_face.vecPt2 = oTrafoFunc(s_face.vecPt2)
            s_face.vecPt3 = oTrafoFunc(s_face.vecPt3)
            s_face.vecPt4 = oTrafoFunc(s_face.vecPt4)

            s_face.vecCentre = as_vec3(as_np3(s_face.vecPt1) + 0.5 * (as_np3(s_face.vecPt3) - as_np3(s_face.vecPt1)))
            s_face.vecLongAxis = as_vec3(normalized(as_np3(s_face.vecPt3) - as_np3(s_face.vecPt1)))
            s_face.vecShortAxis = as_vec3(normalized(as_np3(s_face.vecPt4) - as_np3(s_face.vecPt2)))

        self.m_vecRoundedCentre = None
        self.vecGetRoundedCentre()

    def vecGetRoundedCentre(self) -> Vec3:
        if self.m_vecRoundedCentre is None:
            a_unique_vertices: list[Vec3] = []
            for s_face in self.m_aFaces:
                a_face_vertices = [s_face.vecPt1, s_face.vecPt2, s_face.vecPt3, s_face.vecPt4]
                for vec in a_face_vertices:
                    vec_rounded = (
                        round(float(vec[0]), 4),
                        round(float(vec[1]), 4),
                        round(float(vec[2]), 4),
                    )
                    if vec_rounded not in a_unique_vertices:
                        a_unique_vertices.append(vec_rounded)

            vec_centre = sum((as_np3(vec) for vec in a_unique_vertices), start=as_np3((0.0, 0.0, 0.0)))
            vec_centre /= float(len(a_unique_vertices))
            self.m_vecRoundedCentre = (
                round(float(vec_centre[0]), 4),
                round(float(vec_centre[1]), 4),
                round(float(vec_centre[2]), 4),
            )
        return self.m_vecRoundedCentre

    def nGetNumberOfFaces(self) -> int:
        return len(self.m_aFaces)

    def Preview(self, ePreviewFace: "QuasiTile.EPreviewFace") -> None:
        msh_tile = Mesh()
        vec_centre = as_np3(self.vecGetRoundedCentre())
        for s_face in self.m_aFaces:
            vec_normal = np.cross(
                normalized(as_np3(s_face.vecPt3) - as_np3(s_face.vecPt1)),
                normalized(as_np3(s_face.vecPt4) - as_np3(s_face.vecPt2)),
            )
            if VecOperations.bCheckAlignment(as_vec3(vec_normal), as_vec3(as_np3(s_face.vecCentre) - vec_centre)):
                MeshUtility.AddQuad(msh_tile, s_face.vecPt1, s_face.vecPt2, s_face.vecPt3, s_face.vecPt4)
            else:
                MeshUtility.AddQuad(msh_tile, s_face.vecPt1, s_face.vecPt4, s_face.vecPt3, s_face.vecPt2)

            if ePreviewFace == self.EPreviewFace.AXIS:
                s_face.Preview(False)
            elif ePreviewFace == self.EPreviewFace.CONNECTOR:
                s_face.Preview(True)

        Sh.PreviewMesh(msh_tile, self.m_clr, 0.9)

    def oGetConnectorFrame(self, iFaceIndex: int, iIndex: int = 0) -> LocalFrame:
        s_face = self.m_aFaces[iFaceIndex]

        vec_pos = s_face.vecCentre
        vec_local_x = as_vec3(normalized(as_np3(s_face.vecPt3) - as_np3(s_face.vecPt1)))
        vec_local_y = as_vec3(normalized(as_np3(s_face.vecPt4) - as_np3(s_face.vecPt2)))
        vec_local_z = as_vec3(np.cross(as_np3(vec_local_x), as_np3(vec_local_y)))
        vec_local_z = as_vec3(VecOperations.vecFlipForAlignment(vec_local_z, as_vec3(as_np3(vec_pos) - as_np3(self.vecGetRoundedCentre()))))

        if iIndex != 0 and s_face.eConnector == IcosehedralFace.EConnector.LINE:
            vec_local_x = as_vec3(-as_np3(vec_local_x))

        return LocalFrame(vec_pos, vec_local_z, vec_local_x)

    def AttachToOtherQuasiTile(self, iThisFaceIndex: int, oOtherTile: "QuasiTile", iOtherFaceIndex: int, bSwitch: bool = False) -> None:
        n_this_faces = self.nGetNumberOfFaces()
        n_other_faces = oOtherTile.nGetNumberOfFaces()

        if iThisFaceIndex < 0 or iThisFaceIndex >= n_this_faces:
            raise QuasiTile.ThisFaceNotFoundException("This face index exceeds number of faces on this quasi tile.")
        if iOtherFaceIndex < 0 or iOtherFaceIndex >= n_other_faces:
            raise QuasiTile.OtherFaceNotFoundException("Other face index exceeds number of faces on other quasi tile.")

        if self.m_aFaces[iThisFaceIndex].eConnector != oOtherTile.m_aFaces[iOtherFaceIndex].eConnector:
            raise QuasiTile.ConnectorMismatchException("Connector types do not match.")

        i_switch = 1 if bSwitch else 0
        o_other_connector_frame = oOtherTile.oGetConnectorFrame(iOtherFaceIndex, i_switch)
        o_this_current_face_frame = LocalFrame.oGetInvertFrame(self.oGetConnectorFrame(iThisFaceIndex), True, False)

        for s_face in self.m_aFaces:
            s_face.vecPt1 = VecOperations.vecExpressPointInFrame(o_this_current_face_frame, s_face.vecPt1)
            s_face.vecPt2 = VecOperations.vecExpressPointInFrame(o_this_current_face_frame, s_face.vecPt2)
            s_face.vecPt3 = VecOperations.vecExpressPointInFrame(o_this_current_face_frame, s_face.vecPt3)
            s_face.vecPt4 = VecOperations.vecExpressPointInFrame(o_this_current_face_frame, s_face.vecPt4)

            s_face.vecPt1 = VecOperations.vecTranslatePointOntoFrame(o_other_connector_frame, s_face.vecPt1)
            s_face.vecPt2 = VecOperations.vecTranslatePointOntoFrame(o_other_connector_frame, s_face.vecPt2)
            s_face.vecPt3 = VecOperations.vecTranslatePointOntoFrame(o_other_connector_frame, s_face.vecPt3)
            s_face.vecPt4 = VecOperations.vecTranslatePointOntoFrame(o_other_connector_frame, s_face.vecPt4)

            s_face.vecCentre = as_vec3(as_np3(s_face.vecPt1) + 0.5 * (as_np3(s_face.vecPt3) - as_np3(s_face.vecPt1)))
            s_face.vecLongAxis = as_vec3(normalized(as_np3(s_face.vecPt3) - as_np3(s_face.vecPt1)))
            s_face.vecShortAxis = as_vec3(normalized(as_np3(s_face.vecPt4) - as_np3(s_face.vecPt2)))

        self.m_vecRoundedCentre = None
        self.vecGetRoundedCentre()

    class ThisFaceNotFoundException(Exception):
        pass

    class OtherFaceNotFoundException(Exception):
        pass

    class ConnectorMismatchException(Exception):
        pass


__all__ = ["QuasiTile"]



from .robinson_triangle import RobinsonTriangle
from .rhombic_tile import ISubDTile, LargeRhombicTile, RhombicTile, SmallRhombicTile
from .penrose_pattern import PenrosePattern

__all__ = [
    "ISubDTile",
    "RhombicTile",
    "SmallRhombicTile",
    "LargeRhombicTile",
    "RobinsonTriangle",
    "PenrosePattern",
]




import numpy as np
from dataclasses import dataclass

from ..utils import VecOperations

from .._types import Vec3, Vector3Like, as_np3, as_vec3, normalized

@dataclass
class LocalFrame:
    _position: np.ndarray
    _local_x: np.ndarray
    _local_y: np.ndarray
    _local_z: np.ndarray

    def __init__(
        self,
        vecPos: Vector3Like = (0.0, 0.0, 0.0),
        vecLocalZ: Vector3Like = (0.0, 0.0, 1.0),
        vecLocalX: Vector3Like | None = None,
    ) -> None:
        pos = as_np3(vecPos)
        local_z = normalized(vecLocalZ)
        if float(np.linalg.norm(local_z)) <= 1e-12:
            raise ValueError("Local Z Coordinate has a length of Zero!")
        if vecLocalX is None:
            local_x = VecOperations.vecGetOrthogonalDir(local_z)
        else:
            local_x = normalized(vecLocalX)
            if float(np.linalg.norm(local_x)) <= 1e-12:
                raise ValueError("Local X Coordinate has a length of Zero!")
        local_y = LocalFrame.vecBuildLocalY(local_z, local_x)

        self._position = pos
        self._local_x = normalized(local_x)
        self._local_y = normalized(local_y)
        self._local_z = normalized(local_z)

    def vecGetPosition(self) -> Vec3:
        return as_vec3(self._position)

    def vec_get_position(self) -> Vec3:
        return self.vecGetPosition()

    def vecGetLocalX(self) -> Vec3:
        return as_vec3(self._local_x)

    def vec_get_local_x(self) -> Vec3:
        return self.vecGetLocalX()

    def vecGetLocalY(self) -> Vec3:
        return as_vec3(self._local_y)

    def vec_get_local_y(self) -> Vec3:
        return self.vecGetLocalY()

    def vecGetLocalZ(self) -> Vec3:
        return as_vec3(self._local_z)

    def vec_get_local_z(self) -> Vec3:
        return self.vecGetLocalZ()

    def oInvert(self, bMirrorZ: bool, bMirrorX: bool) -> "LocalFrame":
        z = -self._local_z if bMirrorZ else self._local_z
        x = -self._local_x if bMirrorX else self._local_x
        return LocalFrame(self._position, z, x)

    def oTranslate(self, vecTranslate: Vector3Like) -> "LocalFrame":
        return LocalFrame(self._position + as_np3(vecTranslate), self._local_z, self._local_x)

    def oRotate(self, dPhi: float, vecAxis: Vector3Like) -> "LocalFrame":
        x = VecOperations.vecRotateAroundAxis(self._local_x, dPhi, vecAxis)
        z = VecOperations.vecRotateAroundAxis(self._local_z, dPhi, vecAxis)
        return LocalFrame(self._position, z, x)

    @staticmethod
    def oGetInvertFrame(oFrame: "LocalFrame", bMirrorZ: bool, bMirrorX: bool) -> "LocalFrame":
        return oFrame.oInvert(bMirrorZ, bMirrorX)

    @staticmethod
    def oGetTranslatedFrame(oFrame: "LocalFrame", vecTranslate: Vector3Like) -> "LocalFrame":
        return oFrame.oTranslate(vecTranslate)

    @staticmethod
    def oGetRotatedFrame(oFrame: "LocalFrame", dPhi: float, vecAxis: Vector3Like) -> "LocalFrame":
        return oFrame.oRotate(dPhi, vecAxis)

    @staticmethod
    def vecBuildLocalY(vecLocalZ: Vector3Like, vecLocalX: Vector3Like) -> Vec3:
        z = as_np3(vecLocalZ)
        x = as_np3(vecLocalX)
        return as_vec3(np.cross(z, x))





from __future__ import annotations

import numpy as np

from picogk import BBox3, Library
from shape_kernel import BaseBox, Cp, LocalFrame, Sh, Uf
from shape_kernel._types import Vec3, Vector3Like, as_np3, as_vec3


class RandomDeformationField:
	def __init__(self, oBoundingBox: BBox3, fResolution: float, fMinValue: float, fMaxValue: float) -> None:
		self.m_oBBox = oBoundingBox
		fXSize, fYSize, fZSize = self.m_oBBox.vecSize()
		self.m_nXSamples = int(fXSize / float(fResolution)) + 1
		self.m_nYSamples = int(fYSize / float(fResolution)) + 1
		self.m_nZSamples = int(fZSize / float(fResolution)) + 1

		self.m_aGridPoints = np.zeros((self.m_nXSamples + 1, self.m_nYSamples + 1, self.m_nZSamples + 1, 3), dtype=np.float64)
		self.m_aNoiseVectors = np.zeros_like(self.m_aGridPoints)

		self._preview_field(fXSize, fYSize, fZSize)

		for iX in range(0, self.m_nXSamples + 1):
			for iY in range(0, self.m_nYSamples + 1):
				for iZ in range(0, self.m_nZSamples + 1):
					fX = fXSize / float(self.m_nXSamples) * float(iX) + float(self.m_oBBox.vecMin.x)
					fY = fYSize / float(self.m_nYSamples) * float(iY) + float(self.m_oBBox.vecMin.y)
					fZ = fZSize / float(self.m_nZSamples) * float(iZ) + float(self.m_oBBox.vecMin.z)
					vecPt = np.array([fX, fY, fZ], dtype=np.float64)
					self.m_aGridPoints[iX, iY, iZ] = vecPt

					if Library.oViewer() is not None:
						Sh.PreviewPoint(as_vec3(vecPt), 1.0, Cp().clrRandom())

					self.m_aNoiseVectors[iX, iY, iZ] = np.array(
						[
							Uf.fGetRandomLinear(fMinValue, fMaxValue),
							Uf.fGetRandomLinear(fMinValue, fMaxValue),
							Uf.fGetRandomLinear(fMinValue, fMaxValue),
						],
						dtype=np.float64,
					)

	def _preview_field(self, fXSize: float, fYSize: float, fZSize: float) -> None:
		if Library.oViewer() is None:
			return

		Sh.PreviewPoint((0.0, 0.0, 0.0), 0.5, Cp.clrRed)
		Sh.PreviewBoxWireframe(self.m_oBBox, Cp.clrRed)

		for iX in range(1, self.m_nXSamples + 1):
			for iY in range(1, self.m_nYSamples + 1):
				for iZ in range(1, self.m_nZSamples + 1):
					fMinX = float(self.m_oBBox.vecMin.x) + 1.0 / float(self.m_nXSamples) * (float(iX) - 1.0) * fXSize
					fMaxX = float(self.m_oBBox.vecMin.x) + 1.0 / float(self.m_nXSamples) * float(iX) * fXSize
					fMinY = float(self.m_oBBox.vecMin.y) + 1.0 / float(self.m_nYSamples) * (float(iY) - 1.0) * fYSize
					fMaxY = float(self.m_oBBox.vecMin.y) + 1.0 / float(self.m_nYSamples) * float(iY) * fYSize
					fMinZ = float(self.m_oBBox.vecMin.z) + 1.0 / float(self.m_nZSamples) * (float(iZ) - 1.0) * fZSize
					fMaxZ = float(self.m_oBBox.vecMin.z) + 1.0 / float(self.m_nZSamples) * float(iZ) * fZSize

					vecFramePos = (0.5 * (fMinX + fMaxX), 0.5 * (fMinY + fMaxY), fMinZ)
					dX = fMaxX - fMinX
					dY = fMaxY - fMinY
					dZ = fMaxZ - fMinZ
					oSubCube = BaseBox.from_frame(LocalFrame(vecFramePos), fLength=dZ, fWidth=dX, fDepth=dY)
					Sh.PreviewBoxWireframe(oSubCube, Cp.clrBlack)

	def vecGetData(self, vecPt: Vector3Like) -> Vec3:
		pt = as_np3(vecPt)
		fX = min(max(float(pt[0]), float(self.m_oBBox.vecMin.x)), float(self.m_oBBox.vecMax.x))
		fY = min(max(float(pt[1]), float(self.m_oBBox.vecMin.y)), float(self.m_oBBox.vecMax.y))
		fZ = min(max(float(pt[2]), float(self.m_oBBox.vecMin.z)), float(self.m_oBBox.vecMax.z))

		dX = float(self.m_oBBox.vecSize()[0]) / float(self.m_nXSamples)
		dY = float(self.m_oBBox.vecSize()[1]) / float(self.m_nYSamples)
		dZ = float(self.m_oBBox.vecSize()[2]) / float(self.m_nZSamples)

		iLowerX = int((fX - float(self.m_oBBox.vecMin.x)) / dX)
		iLowerY = int((fY - float(self.m_oBBox.vecMin.y)) / dY)
		iLowerZ = int((fZ - float(self.m_oBBox.vecMin.z)) / dZ)
		iLowerX = min(max(iLowerX, 0), self.m_nXSamples - 1)
		iLowerY = min(max(iLowerY, 0), self.m_nYSamples - 1)
		iLowerZ = min(max(iLowerZ, 0), self.m_nZSamples - 1)
		iUpperX = iLowerX + 1
		iUpperY = iLowerY + 1
		iUpperZ = iLowerZ + 1

		vecPt000 = self.m_aGridPoints[iLowerX, iLowerY, iLowerZ]

		vecDir000 = self.m_aNoiseVectors[iLowerX, iLowerY, iLowerZ]
		vecDir100 = self.m_aNoiseVectors[iUpperX, iLowerY, iLowerZ]
		vecDir110 = self.m_aNoiseVectors[iUpperX, iUpperY, iLowerZ]
		vecDir010 = self.m_aNoiseVectors[iLowerX, iUpperY, iLowerZ]
		vecDir001 = self.m_aNoiseVectors[iLowerX, iLowerY, iUpperZ]
		vecDir101 = self.m_aNoiseVectors[iUpperX, iLowerY, iUpperZ]
		vecDir111 = self.m_aNoiseVectors[iUpperX, iUpperY, iUpperZ]
		vecDir011 = self.m_aNoiseVectors[iLowerX, iUpperY, iUpperZ]

		fXRatio = (fX - float(vecPt000[0])) / dX
		fYRatio = (fY - float(vecPt000[1])) / dY
		fZRatio = (fZ - float(vecPt000[2])) / dZ

		vecDir00 = self.vecGetInter(vecDir000, vecDir100, fXRatio)
		vecDir10 = self.vecGetInter(vecDir010, vecDir110, fXRatio)
		vecDir01 = self.vecGetInter(vecDir001, vecDir101, fXRatio)
		vecDir11 = self.vecGetInter(vecDir011, vecDir111, fXRatio)
		vecDir0 = self.vecGetInter(vecDir00, vecDir10, fYRatio)
		vecDir1 = self.vecGetInter(vecDir01, vecDir11, fYRatio)
		vecDir = self.vecGetInter(vecDir0, vecDir1, fZRatio)
		return as_vec3(vecDir)

	def vecGetInter(self, vecMin: np.ndarray, vecMax: np.ndarray, fRatio: float) -> np.ndarray:
		ratio = min(max(float(fRatio), 0.0), 1.0)
		return vecMin + ratio * (vecMax - vecMin)


__all__ = ["RandomDeformationField"]



from enum import IntEnum
import time
from typing import Callable, Protocol

class IAction(Protocol):
    def bAnimate(self, fTimeStamp: float) -> bool:
        ...


class Animation:
    class EType(IntEnum):
        ONCE = 0
        REPEAT = 1
        WIGGLE = 2

    def __init__(self, action: Callable[[float], None], duration_seconds: float, eType: "Animation.EType" = EType.ONCE) -> None:
        if duration_seconds <= 0.0:
            raise ValueError("duration_seconds must be > 0")
        self._action = action
        self._duration = float(duration_seconds)
        self._type = eType
        self._start_time = time.perf_counter()

    def bAnimate(self, fTimeStamp: float) -> bool:
        dt = max(0.0, float(fTimeStamp) - self._start_time)
        finished = dt >= self._duration
        t = dt / self._duration
        if self._type == Animation.EType.ONCE:
            if finished:
                t = 1.0
        elif self._type == Animation.EType.REPEAT:
            t = t % 1.0
            finished = False
        else:
            oscillation = t % 2.0
            t = oscillation if oscillation <= 1.0 else 2.0 - oscillation
            finished = False
        self._action(float(t))
        return not finished

    def End(self) -> None:
        self._action(1.0)


class AnimationQueue:
    def __init__(self) -> None:
        self._queue: list[IAction] = []

    def Clear(self) -> None:
        for action in self._queue:
            end_fn = getattr(action, "End", None)
            if callable(end_fn):
                end_fn()
        self._queue.clear()

    def Add(self, action: IAction) -> None:
        self._queue.append(action)

    def bIsIdle(self) -> bool:
        return len(self._queue) == 0

    def bPulse(self) -> bool:
        if not self._queue:
            return False
        now = time.perf_counter()
        to_remove: list[int] = []
        for i, action in enumerate(self._queue):
            if not action.bAnimate(now):
                to_remove.append(i)
        for idx in reversed(to_remove):
            self._queue.pop(idx)
        return len(self._queue) > 0





from pathlib import Path
import struct

from _common.types import clamp01

from .image import Image, ImageColor, ImageGrayScale


class TgaIo:
    @staticmethod
    def bYAxisFlipped(image_desc: int) -> bool:
        return (int(image_desc) & 0x20) != 0

    @staticmethod
    def SaveTga(file_path: str | Path, img: Image) -> None:
        width = img.nWidth
        height = img.nHeight
        is_color = img.eType == Image.EType.COLOR
        image_type = 2 if is_color else 3
        pixel_depth = 24 if is_color else 8
        header = struct.pack(
            "<BBBHHBHHHHBB",
            0,
            0,
            image_type,
            0,
            0,
            0,
            0,
            0,
            width,
            height,
            pixel_depth,
            32,
        )
        with open(file_path, "wb") as handle:
            handle.write(header)
            for y in range(height):
                for x in range(width):
                    if is_color:
                        r, g, b, _ = img.clrValue(x, y)
                        handle.write(bytes((int(clamp01(b) * 255), int(clamp01(g) * 255), int(clamp01(r) * 255))))
                    else:
                        handle.write(bytes((img.byGetValue(x, y),)))

    @staticmethod
    def GetFileInfo(file_path: str | Path) -> tuple[Image.EType, int, int]:
        with open(file_path, "rb") as handle:
            header = handle.read(18)
        if len(header) != 18:
            raise ValueError("TGA header too short")
        _, _, image_type, _, _, _, _, _, width, height, _, _ = struct.unpack("<BBBHHBHHHHBB", header)
        eType = Image.EType.COLOR if image_type == 2 else Image.EType.GRAY
        return (eType, int(width), int(height))

    @staticmethod
    def LoadTga(file_path: str | Path) -> Image:
        with open(file_path, "rb") as handle:
            header = handle.read(18)
            if len(header) != 18:
                raise ValueError("TGA header too short")
            _, _, image_type, _, _, _, _, _, width, height, pixel_depth, image_desc = struct.unpack("<BBBHHBHHHHBB", header)
            if image_type == 2 and pixel_depth == 24:
                img: Image = ImageColor(width, height)
                for y in range(height):
                    iy = y if TgaIo.bYAxisFlipped(image_desc) else (height - y - 1)
                    for x in range(width):
                        bgr = handle.read(3)
                        if len(bgr) != 3:
                            raise ValueError("Unexpected EOF in TGA payload")
                        b, g, r = bgr[0], bgr[1], bgr[2]
                        img.SetValue(x, iy, (r / 255.0, g / 255.0, b / 255.0, 1.0))
                return img
            if image_type == 3 and pixel_depth == 8:
                img = ImageGrayScale(width, height)
                for y in range(height):
                    iy = y if TgaIo.bYAxisFlipped(image_desc) else (height - y - 1)
                    for x in range(width):
                        value = handle.read(1)
                        if len(value) != 1:
                            raise ValueError("Unexpected EOF in TGA payload")
                        img.SetValue(x, iy, value[0] / 255.0)
                return img
        raise ValueError("Unsupported TGA format")





from __future__ import annotations

import random

from picogk import Voxels
from shape_kernel._types import Vec3, as_vec3

from .._helpers import _coerce_bbox3
from ..unit_cells import CuboidCell, IUnitCell
from .i_cell_array import ICellArray


class RegularCellArray(ICellArray):
	def __init__(self, oVoxels: Voxels, dX: float, dY: float, dZ: float, fNoiseLevel: float = 0.0) -> None:
		self.m_dX = float(dX)
		self.m_dY = float(dY)
		self.m_dZ = float(dZ)
		self.m_fNoiseLevel = min(max(abs(float(fNoiseLevel)), 0.0), 0.3)
		self.m_oRandom = random.Random()
		self.m_oVoxels = oVoxels
		_, oBBox = self.m_oVoxels.CalculateProperties()
		bbox = _coerce_bbox3(oBBox)

		self.m_aUnitCells: list[IUnitCell] = []
		fX = float(bbox.vecMin.x - 0.5 * self.m_dX)
		fXEnd = float(bbox.vecMax.x + 0.5 * self.m_dX)
		while fX <= fXEnd:
			fY = float(bbox.vecMin.y - 0.5 * self.m_dY)
			fYEnd = float(bbox.vecMax.y + 0.5 * self.m_dY)
			while fY <= fYEnd:
				fZ = float(bbox.vecMin.z - 0.5 * self.m_dZ)
				fZEnd = float(bbox.vecMax.z + 0.5 * self.m_dZ)
				while fZ <= fZEnd:
					self.m_aUnitCells.append(
						CuboidCell(
							self.vecGetCorner(fX, fY, fZ),
							self.vecGetCorner(fX, fY + self.m_dY, fZ),
							self.vecGetCorner(fX + self.m_dX, fY + self.m_dY, fZ),
							self.vecGetCorner(fX + self.m_dX, fY, fZ),
							self.vecGetCorner(fX, fY, fZ + self.m_dZ),
							self.vecGetCorner(fX, fY + self.m_dY, fZ + self.m_dZ),
							self.vecGetCorner(fX + self.m_dX, fY + self.m_dY, fZ + self.m_dZ),
							self.vecGetCorner(fX + self.m_dX, fY, fZ + self.m_dZ),
						)
					)
					fZ += self.m_dZ
				fY += self.m_dY
			fX += self.m_dX

	def vecGetCorner(self, fX: float, fY: float, fZ: float) -> Vec3:
		iX = int(fX * 1000.0)
		iY = int(fY * 1000.0)
		iZ = int(fZ * 1000.0)
		iIndex = iX * iY * iZ
		self.m_oRandom = random.Random(iIndex)
		vecNoise = self.vecGetNoise()
		return as_vec3((fX + vecNoise[0], fY + vecNoise[1], fZ + vecNoise[2]))

	def vecGetNoise(self) -> Vec3:
		return (
			self.fGetRandomLinear(-self.m_fNoiseLevel * self.m_dX, self.m_fNoiseLevel * self.m_dX),
			self.fGetRandomLinear(-self.m_fNoiseLevel * self.m_dY, self.m_fNoiseLevel * self.m_dY),
			self.fGetRandomLinear(-self.m_fNoiseLevel * self.m_dZ, self.m_fNoiseLevel * self.m_dZ),
		)

	def fGetRandomLinear(self, fMin: float, fMax: float, oRandom: random.Random | None = None) -> float:
		rng = self.m_oRandom if oRandom is None else oRandom
		return float(fMin) + (float(fMax) - float(fMin)) * rng.random()

	def aGetUnitCells(self) -> list[IUnitCell]:
		return self.m_aUnitCells


__all__ = ["RegularCellArray"]


from ._viewer import (
    AnimGroupMatrixRotate,
    AnimViewRotate,
    IKeyHandler,
    IViewerAction,
    KeyAction,
    KeyHandler,
    RotateToNextRoundAngleAction,
    VedoViewer,
    go,
)

from ._viewer_protocol import IViewer

from ._core import (
    IImplicit,
    IBoundedImplicit,
    Config,
    BBox3,
    FieldMetadata,
    Lattice,
    Library,
    Mesh,
    OpenVdbFile,
    PolyLine,
    ScalarField,
    VectorField,
    Voxels,
    BBox2,
    ActiveVoxelCounterScalar,
    Animation,
    AnimationQueue,
    Cli,
    CliIo,
    CsvTable,
    ESliceMode,
    EStlUnit,
    ImageRgba32,
    ImageRgb24,
    Easing,
    FieldUtils,
    Image,
    ImageBW,
    ImageColor,
    ImageGrayScale,
    MeshIo,
    MeshMath,
    PolyContour,
    PolySlice,
    PolySliceStack,
    SdfVisualizer,
    AddVectorFieldToViewer,
    LogFile,
    ImplicitMesh,
    ImplicitTriangle,
    TempFolder,
    TgaIo,
    TriangleVoxelization,
    Utils,
    Vector3Ext,
    VectorFieldMerge,
    SurfaceNormalFieldExtractor,
)
from ._errors import PicoGKDisposedError, PicoGKError, PicoGKInvalidHandleError, PicoGKLoadError

__all__ = [
    "IImplicit",
    "IBoundedImplicit",
    "PicoGKError",
    "Easing",
    "PicoGKLoadError",
    "PicoGKDisposedError",
    "PicoGKInvalidHandleError",
    "Config",
    "Library",
    "KeyAction",
    "KeyHandler",
    "IKeyHandler",
    "IViewerAction",
    "RotateToNextRoundAngleAction",
    "AnimGroupMatrixRotate",
    "AnimViewRotate",
    "Mesh",
    "MeshIo",
    "BBox2",
    "BBox3",
    "EStlUnit",
    "ESliceMode",
    "MeshMath",
    "Lattice",
    "Voxels",
    "ScalarField",
    "VectorField",
    "FieldUtils",
    "ActiveVoxelCounterScalar",
    "SurfaceNormalFieldExtractor",
    "VectorFieldMerge",
    "SdfVisualizer",
    "AddVectorFieldToViewer",
    "Image",
    "ImageBW",
    "ImageGrayScale",
    "ImageColor",
    "TgaIo",
    "PolyContour",
    "PolySlice",
    "PolySliceStack",
    "Cli",
    "CliIo",
    "CsvTable",
    "Animation",
    "AnimationQueue",
    "ImplicitTriangle",
    "ImplicitMesh",
    "TriangleVoxelization",
    "Utils",
    "Vector3Ext",
    "TempFolder",
    "LogFile",
    "FieldMetadata",
    "OpenVdbFile",
    "PolyLine",
    "IViewer",
    "VedoViewer",
    "go",
    "ImageRgb24",
    "ImageRgba32",
]



from .conformal_cell_array import ConformalCellArray, ConformalShowcaseShapes
from .i_cell_array import ICellArray
from .regular_cell_array import RegularCellArray
from .regular_unit_cell import RegularUnitCell

__all__ = ["ICellArray", "ConformalCellArray", "ConformalShowcaseShapes", "RegularCellArray", "RegularUnitCell"]


from __future__ import annotations

from picogk import Lattice
from shape_kernel._types import as_np3, as_vec3

from .._helpers import _add_sampled_beam
from ..beam_thickness import IBeamThickness
from ..unit_cells import IUnitCell
from .i_lattice_type import ILatticeType


class OctahedronLattice(ILatticeType):
	def AddCell(self, oLattice: Lattice, xCell: IUnitCell, xBeamThickness: IBeamThickness, nSubSamples: int = 2) -> None:
		aCornerPoints = xCell.aGetCornerPoints()
		if len(aCornerPoints) != 8:
			raise ValueError("Octahedron Lattice only supports Unit Cells with 8 Corners.")

		pts = [as_np3(point) for point in aCornerPoints]
		vecLowerFace = as_vec3(0.25 * (pts[0] + pts[1] + pts[2] + pts[3]))
		vecUpperFace = as_vec3(0.25 * (pts[4] + pts[5] + pts[6] + pts[7]))
		vecForwardFace = as_vec3(0.25 * (pts[4] + pts[5] + pts[0] + pts[1]))
		vecRightFace = as_vec3(0.25 * (pts[6] + pts[5] + pts[2] + pts[1]))
		vecBackwardFace = as_vec3(0.25 * (pts[6] + pts[7] + pts[2] + pts[3]))
		vecLeftFace = as_vec3(0.25 * (pts[4] + pts[7] + pts[0] + pts[3]))

		for vecPt1, vecPt2 in (
			(vecLowerFace, vecRightFace),
			(vecLowerFace, vecLeftFace),
			(vecLowerFace, vecForwardFace),
			(vecLowerFace, vecBackwardFace),
			(vecUpperFace, vecRightFace),
			(vecUpperFace, vecLeftFace),
			(vecUpperFace, vecForwardFace),
			(vecUpperFace, vecBackwardFace),
			(vecForwardFace, vecRightFace),
			(vecForwardFace, vecLeftFace),
			(vecBackwardFace, vecRightFace),
			(vecBackwardFace, vecLeftFace),
		):
			_add_sampled_beam(oLattice, vecPt1, vecPt2, xBeamThickness, nSubSamples)


__all__ = ["OctahedronLattice"]


from __future__ import annotations

from enum import IntEnum
import struct
from typing import TYPE_CHECKING

from _common.types import Vector3Like
import numpy as np
from pathlib import Path

if TYPE_CHECKING:
    from .mesh import Mesh


class EStlUnit(IntEnum):
    AUTO = 0
    MM = 1
    CM = 2
    M = 3
    FT = 4
    IN = 5


def _stl_unit_scale_to_mm(unit: EStlUnit) -> float:
    return {
        EStlUnit.MM: 1.0,
        EStlUnit.CM: 10.0,
        EStlUnit.M: 1000.0,
        EStlUnit.FT: 304.8,
        EStlUnit.IN: 25.4,
    }.get(unit, 1.0)


class MeshIo:
    @staticmethod
    def mshFromStlFile(
        strFilePath: str,
        eLoadUnit: EStlUnit = EStlUnit.AUTO,
        fPostScale: float = 1.0,
        vecPostOffsetMM: Vector3Like | None = None,
    ) -> Mesh:
        from .mesh import Mesh

        path = Path(strFilePath)
        raw = path.read_bytes()
        if len(raw) < 84:
            raise ValueError("Failed to read STL file, header too short")
        header = raw[:80].decode("ascii", errors="ignore")
        offset = np.array(
            vecPostOffsetMM if vecPostOffsetMM is not None else (0.0, 0.0, 0.0),
            dtype=np.float64,
        )

        load_unit = eLoadUnit
        if load_unit == EStlUnit.AUTO:
            h = header.lower()
            if "units=cm" in h:
                load_unit = EStlUnit.CM
            elif "units= m" in h:
                load_unit = EStlUnit.M
            elif "units=ft" in h:
                load_unit = EStlUnit.FT
            elif "units=in" in h:
                load_unit = EStlUnit.IN
            else:
                load_unit = EStlUnit.MM

        scale = _stl_unit_scale_to_mm(load_unit) * float(fPostScale)
        msh = Mesh()
        msh.m_strLoadHeaderData = header.strip()
        msh.m_eLoadUnits = load_unit

        is_ascii = raw.startswith(b"solid") and (b"vertex" in raw[:2048])
        if is_ascii:
            text = raw.decode("ascii", errors="ignore")
            verts: list[tuple[float, float, float]] = []
            for line in text.splitlines():
                line = line.strip().lower()
                if line.startswith("vertex "):
                    parts = line.split()
                    if len(parts) >= 4:
                        v = np.array(
                            (float(parts[1]), float(parts[2]), float(parts[3])),
                            dtype=np.float64,
                        )
                        v = (v * scale) + offset
                        verts.append((float(v[0]), float(v[1]), float(v[2])))
                        if len(verts) == 3:
                            msh.nAddTriangle(verts[0], verts[1], verts[2])
                            verts.clear()
        else:
            tri_count = struct.unpack("<I", raw[80:84])[0]
            cursor = 84
            for _ in range(tri_count):
                if cursor + 50 > len(raw):
                    break
                data = raw[cursor : cursor + 50]
                cursor += 50
                v0 = np.array(struct.unpack("<3f", data[12:24]), dtype=np.float64)
                v1 = np.array(struct.unpack("<3f", data[24:36]), dtype=np.float64)
                v2 = np.array(struct.unpack("<3f", data[36:48]), dtype=np.float64)
                v0 = (v0 * scale) + offset
                v1 = (v1 * scale) + offset
                v2 = (v2 * scale) + offset
                msh.nAddTriangle(
                    tuple(v0.tolist()), tuple(v1.tolist()), tuple(v2.tolist())
                )

        if msh.nTriangleCount() == 0:
            msh.close()
            raise ValueError(
                "Imported STL mesh is empty (zero triangles), failed to load"
            )
        return msh

    @staticmethod
    def SaveToStlFile(
        mesh: Mesh,
        strFilePath: str,
        eUnit: EStlUnit = EStlUnit.AUTO,
        vecOffsetMM: Vector3Like | None = None,
        fScale: float = 1.0,
    ) -> None:
        unit = eUnit
        if unit == EStlUnit.AUTO:
            unit = getattr(mesh, "m_eLoadUnits", EStlUnit.MM)
        unit_tag = {
            EStlUnit.CM: "UNITS=cm",
            EStlUnit.M: "UNITS= m",
            EStlUnit.FT: "UNITS=ft",
            EStlUnit.IN: "UNITS=in",
        }.get(unit, "UNITS=mm")
        header = f"PicoGK {unit_tag}".ljust(80, " ").encode("ascii", errors="ignore")
        conv = 1.0 / _stl_unit_scale_to_mm(unit)
        offset = np.array(
            vecOffsetMM if vecOffsetMM is not None else (0.0, 0.0, 0.0),
            dtype=np.float64,
        )

        with open(strFilePath, "wb") as handle:
            handle.write(header)
            handle.write(struct.pack("<I", mesh.nTriangleCount()))
            for i in range(mesh.nTriangleCount()):
                a, b, c = mesh.GetTriangle(i)
                va = (np.array(a, dtype=np.float64) + offset) * fScale * conv
                vb = (np.array(b, dtype=np.float64) + offset) * fScale * conv
                vc = (np.array(c, dtype=np.float64) + offset) * fScale * conv
                normal = np.cross(vb - va, vc - va)
                nlen = np.linalg.norm(normal)
                if nlen > 1e-12:
                    normal = normal / nlen
                else:
                    normal = np.array((0.0, 0.0, 1.0), dtype=np.float64)
                handle.write(
                    struct.pack(
                        "<3f", float(normal[0]), float(normal[1]), float(normal[2])
                    )
                )
                handle.write(
                    struct.pack("<3f", float(va[0]), float(va[1]), float(va[2]))
                )
                handle.write(
                    struct.pack("<3f", float(vb[0]), float(vb[1]), float(vb[2]))
                )
                handle.write(
                    struct.pack("<3f", float(vc[0]), float(vc[1]), float(vc[2]))
                )
                handle.write(struct.pack("<H", 0))



"""Leaf module – defines the IViewer structural protocol.

This module has *no* imports from the rest of the picogk package, so it can be
imported by both the ``data`` layer (Library) and the ``viewer`` layer
(VedoViewer) without creating any circular dependency.
"""
from __future__ import annotations

from typing import Any, Protocol, Sequence, runtime_checkable


@runtime_checkable
class IViewer(Protocol):
    """Structural interface that any viewer object passed to Library must satisfy.

    VedoViewer (and any future viewer implementation) satisfies this protocol
    through duck-typing – no explicit inheritance is required.
    """

    def Add(self, xObject: object, nGroupID: int = 0) -> int: ...
    def RequestUpdate(self) -> None: ...
    def RequestScreenShot(self, strScreenShotPath: str) -> None: ...
    def SetGroupVisible(self, nGroupID: int, bVisible: bool) -> None: ...
    def SetGroupStatic(self, nGroupID: int, bStatic: bool) -> None: ...
    def SetGroupMaterial(
        self,
        nGroupID: int,
        clr: Sequence[float],
        fMetallic: float,
        fRoughness: float,
    ) -> None: ...
    def SetGroupMatrix(self, nGroupID: int, mat: Sequence[float]) -> None: ...
    def SetViewAngles(self, fOrbit: float, fElevation: float) -> None: ...
    def AdjustViewAngles(
        self, fOrbitRelative: float, fElevationRelative: float
    ) -> None: ...
    def AddKeyHandler(self, xKeyHandler: Any) -> None: ...
    def AddAnimation(self, oAnim: Any) -> None: ...



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable

from picogk import Voxels

from .._types import Vec3


class BaseShape(ABC):
	fnVertexTransformation = Callable[[Vec3], Vec3]

	def __init__(self) -> None:
		self.m_fnTrafo: BaseShape.fnVertexTransformation = self.vecNoTransform

	def SetTransformation(self, fnTrafo: fnVertexTransformation) -> None:
		self.m_fnTrafo = fnTrafo

	def set_transformation(self, fnTrafo: fnVertexTransformation) -> None:
		self.SetTransformation(fnTrafo)

	@staticmethod
	def vecNoTransform(vecPt: Vec3) -> Vec3:
		return vecPt

	@abstractmethod
	def voxConstruct(self) -> Voxels:
		raise NotImplementedError


__all__ = ["BaseShape"]




from dataclasses import dataclass
from enum import IntEnum
import math
from pathlib import Path
import time
from typing import Sequence

from .slice import PolyContour, PolySlice, PolySliceStack
from .utils import Utils


class Cli:
    class EFormat(IntEnum):
        UseEmptyFirstLayer = 0
        FirstLayerWithContent = 1

    @dataclass
    class Result:
        oSlices: PolySliceStack
        oBBoxFile: tuple[tuple[float, float, float], tuple[float, float, float]]
        bBinary: bool
        fUnitsHeader: float
        b32BitAlign: bool
        nVersion: int
        strHeaderDate: str
        nLayers: int
        strWarnings: str

    @staticmethod
    def _extract_parameter(line: str) -> tuple[bool, str, str]:
        if not line:
            return (False, "", line)
        if line[0] not in "/,":
            return (False, "", line)
        line = line[1:]
        end = len(line)
        for i, ch in enumerate(line):
            if ch in "$/,":
                end = i
                break
        param = line[:end]
        rest = line[end:]
        return (True, param, rest)

    @staticmethod
    def _safe_float(text: str, context: str) -> float:
        try:
            return float(text)
        except Exception as exc:  # noqa: BLE001
            raise ValueError(f"Invalid parameter for {context}: {text}") from exc

    @staticmethod
    def _safe_uint(text: str, context: str) -> int:
        try:
            value = int(text)
        except Exception as exc:  # noqa: BLE001
            raise ValueError(f"Invalid parameter for {context}: {text}") from exc
        if value < 0:
            raise ValueError(f"Invalid parameter for {context}: {text}")
        return value

    @staticmethod
    def WriteSlicesToCliFile(
        oSlices: PolySliceStack | Sequence[PolySlice],
        strFilePath: str | Path,
        eFormat: "Cli.EFormat" = EFormat.FirstLayerWithContent,
        strDate: str = "",
        fUnitsInMM: float = 0.0,
    ) -> None:
        stack = oSlices if isinstance(oSlices, PolySliceStack) else PolySliceStack(oSlices)

        if stack.nCount() < 1 or (math.isinf(stack.oBBox().vecMin[0]) and math.isinf(stack.oBBox().vecMin[1])):
            raise ValueError("No valid slices detected (empty)")

        if fUnitsInMM <= 0.0:
            fUnitsInMM = 1.0
        if not strDate:
            strDate = time.strftime("%Y-%m-%d")

        bbox = stack.oBBox()
        nSliceCount = stack.nCount() + (1 if eFormat == Cli.EFormat.UseEmptyFirstLayer else 0)
        strDim = (
            f"{bbox.vecMin[0]:08.5f},{bbox.vecMin[1]:08.5f},{0.0:08.5f},"
            f"{bbox.vecMax[0]:08.5f},{bbox.vecMax[1]:08.5f},{stack.oSliceAt(stack.nCount() - 1).fZPos():08.5f}"
        )

        with open(strFilePath, "w", encoding="ascii", newline="\n") as handle:
            handle.write("$$HEADERSTART\n")
            handle.write("$$ASCII\n")
            handle.write(f"$$UNITS/{fUnitsInMM}\n")
            handle.write("$$VERSION/200\n")
            handle.write("$$LABEL/1,default\n")
            handle.write(f"$$DATE/{strDate}\n")
            handle.write(f"$$DIMENSION/{strDim}\n")
            handle.write(f"$$LAYERS/{nSliceCount:05d}\n")
            handle.write("$$HEADEREND\n")
            handle.write("$$GEOMETRYSTART\n")

            if eFormat == Cli.EFormat.UseEmptyFirstLayer:
                handle.write("$$LAYER/0.0\n")

            for nLayer in range(stack.nCount()):
                sl = stack.oSliceAt(nLayer)
                handle.write(f"$$LAYER/{(sl.fZPos() / fUnitsInMM):0.5f}\n")

                for nPass in range(3):
                    for nPolyline in range(sl.nContours()):
                        poly = sl.oContourAt(nPolyline)
                        if nPass == 0 and poly.eWinding() != PolyContour.EWinding.COUNTERCLOCKWISE:
                            continue
                        if nPass == 1 and poly.eWinding() != PolyContour.EWinding.CLOCKWISE:
                            continue
                        if nPass == 2 and poly.eWinding() != PolyContour.EWinding.UNKNOWN:
                            continue

                        nWinding = 2
                        if poly.eWinding() == PolyContour.EWinding.CLOCKWISE:
                            nWinding = 0
                        elif poly.eWinding() == PolyContour.EWinding.COUNTERCLOCKWISE:
                            nWinding = 1

                        entries = ["$$POLYLINE/1", str(nWinding), str(poly.nCount())]
                        for nVertex in range(poly.nCount()):
                            vx, vy = poly.vecVertex(nVertex)
                            entries.append(f"{(vx / fUnitsInMM):0.5f}")
                            entries.append(f"{(vy / fUnitsInMM):0.5f}")
                        handle.write(",".join(entries) + "\n")

            handle.write("$$GEOMETRYEND\n")

    @staticmethod
    def oSlicesFromCliFile(strFilePath: str | Path) -> "Cli.Result":
        with open(strFilePath, "r", encoding="ascii", errors="ignore") as handle:
            raw_lines = handle.readlines()

        warnings: list[str] = []
        in_comment = False
        n_label = None
        n_header_layer_numbers = 0
        b_binary = False
        f_units_header = 0.0
        b_32bit_align = False
        n_version = 0
        str_header_date = ""
        bbox_file = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]

        i = 0
        b_header_started = False
        b_header_ended = False
        while i < len(raw_lines) and not b_header_ended:
            line = raw_lines[i]
            i += 1

            while line != "":
                line = line.strip()
                if line.startswith("//"):
                    in_comment = True
                    line = line[2:]

                if in_comment:
                    end = line.find("//")
                    if end == -1:
                        line = ""
                        continue
                    line = line[end + 2 :].strip()
                    in_comment = False

                if not b_header_started:
                    idx = line.find("$$HEADERSTART")
                    if idx == -1:
                        break
                    line = line[idx + len("$$HEADERSTART") :].strip()
                    b_header_started = True
                    continue

                if not line.startswith("$$"):
                    break

                if line.startswith("$$BINARY"):
                    line = line[len("$$BINARY") :]
                    b_binary = True
                    continue
                if line.startswith("$$ASCII"):
                    line = line[len("$$ASCII") :]
                    b_binary = False
                    continue
                if line.startswith("$$ALIGN"):
                    line = line[len("$$ALIGN") :]
                    b_32bit_align = True
                    continue
                if line.startswith("$$UNITS"):
                    line = line[len("$$UNITS") :]
                    ok, param, line = Cli._extract_parameter(line)
                    if not ok:
                        raise ValueError("Missing parameter after $$UNITS")
                    f_units_header = Cli._safe_float(param, "$$UNITS")
                    if f_units_header <= 0.0:
                        raise ValueError(f"Invalid parameter for $$UNITS: {param}")
                    continue
                if line.startswith("$$VERSION"):
                    line = line[len("$$VERSION") :]
                    ok, param, line = Cli._extract_parameter(line)
                    if not ok:
                        raise ValueError("Missing parameter after $$VERSION")
                    n_version = Cli._safe_uint(param, "$$VERSION")
                    continue
                if line.startswith("$$LABEL"):
                    line = line[len("$$LABEL") :]
                    ok, param, line = Cli._extract_parameter(line)
                    if not ok:
                        raise ValueError("Missing parameter after $$LABEL")
                    if n_label is not None:
                        raise NotImplementedError("Currently we do not support multiple labels and objects in one CLI file")
                    n_label = Cli._safe_uint(param, "$$LABEL")
                    ok, _, line = Cli._extract_parameter(line)
                    if not ok:
                        raise ValueError("Missing parameter after $$LABEL (text)")
                    continue
                if line.startswith("$$DATE"):
                    line = line[len("$$DATE") :]
                    ok, param, line = Cli._extract_parameter(line)
                    if not ok:
                        raise ValueError("Missing parameter after $$DATE")
                    str_header_date = param.strip()
                    continue
                if line.startswith("$$DIMENSION"):
                    line = line[len("$$DIMENSION") :]
                    values: list[float] = []
                    for _ in range(6):
                        ok, param, line = Cli._extract_parameter(line)
                        if not ok:
                            raise ValueError("Missing parameter after $$DIMENSION")
                        values.append(Cli._safe_float(param, "$$DIMENSION"))
                    bbox_file[0] = [values[0], values[1], values[2]]
                    bbox_file[1] = [values[3], values[4], values[5]]
                    continue
                if line.startswith("$$LAYERS"):
                    line = line[len("$$LAYERS") :]
                    ok, param, line = Cli._extract_parameter(line)
                    if not ok:
                        raise ValueError("Missing parameter after $$LAYERS")
                    n_header_layer_numbers = Cli._safe_uint(param, "$$LAYERS")
                    continue
                if line.startswith("$$HEADEREND"):
                    b_header_ended = True
                    break

                break

        if not b_header_ended:
            raise ValueError("End of file while searching for valid header")

        if b_binary:
            raise NotImplementedError("Binary CLI Files are not yet supported")

        b_geometry_started = False
        b_geometry_ended = False
        o_current_slice: PolySlice | None = None
        slices: list[PolySlice] = []
        f_prev_z = float("-inf")

        while i < len(raw_lines) and not b_geometry_ended:
            line = raw_lines[i]
            i += 1
            while line != "":
                line = line.strip()
                if line.startswith("//"):
                    in_comment = True
                    line = line[2:]

                if in_comment:
                    end = line.find("//")
                    if end == -1:
                        line = ""
                        continue
                    line = line[end + 2 :].strip()
                    in_comment = False

                if not b_geometry_started:
                    idx = line.find("$$GEOMETRYSTART")
                    if idx == -1:
                        break
                    line = line[idx + len("$$GEOMETRYSTART") :].strip()
                    b_geometry_started = True
                    continue

                if line.startswith("$$LAYER"):
                    line = line[len("$$LAYER") :]
                    ok, param, line = Cli._extract_parameter(line)
                    if not ok:
                        raise ValueError("Missing parameter after $$LAYER")
                    f_z = Cli._safe_float(param, "$$LAYER") * f_units_header

                    if f_prev_z != float("-inf") and f_z < f_prev_z:
                        raise ValueError(f"Z position in current layer is smaller than in previous {param}")

                    if f_z > 0.0:
                        if o_current_slice is not None:
                            slices.append(o_current_slice)
                        o_current_slice = PolySlice(f_z)
                        f_prev_z = f_z
                    continue

                if line.startswith("$$POLYLINE"):
                    if o_current_slice is None:
                        raise ValueError("There should not be contours at z position 0")

                    line = line[len("$$POLYLINE") :]
                    ok, param, line = Cli._extract_parameter(line)
                    if not ok:
                        raise ValueError("Missing parameter after $$POLYLINE")
                    n_id = Cli._safe_uint(param, "$$POLYLINE")

                    if n_label is None:
                        n_label = n_id
                    if n_id != n_label:
                        raise NotImplementedError("We do not support CLI labels and multiple models yet")

                    ok, param, line = Cli._extract_parameter(line)
                    if not ok:
                        raise ValueError("Missing parameter after $$POLYLINE")
                    n_winding = Cli._safe_uint(param, "$$POLYLINE direction")
                    if n_winding == 0:
                        declared = PolyContour.EWinding.CLOCKWISE
                    elif n_winding == 1:
                        declared = PolyContour.EWinding.COUNTERCLOCKWISE
                    elif n_winding == 2:
                        declared = PolyContour.EWinding.UNKNOWN
                    else:
                        raise ValueError(f"Invalid parameter for $$POLYLINE direction: {param}")

                    ok, param, line = Cli._extract_parameter(line)
                    if not ok:
                        raise ValueError("Missing parameter polygon count after $$POLYLINE")
                    n_count = Cli._safe_uint(param, "$$POLYLINE polygon count")

                    vertices: list[tuple[float, float]] = []
                    while n_count > 0:
                        ok, param, line = Cli._extract_parameter(line)
                        if not ok:
                            raise ValueError("Missing vertices in $$POLYLINE")
                        fx = Cli._safe_float(param, "$$POLYLINE vertex X") * f_units_header
                        ok, param, line = Cli._extract_parameter(line)
                        if not ok:
                            raise ValueError("Missing vertices in $$POLYLINE")
                        fy = Cli._safe_float(param, "$$POLYLINE vertex Y") * f_units_header
                        vertices.append((fx, fy))
                        n_count -= 1

                    if len(vertices) < 3:
                        warnings.append(f"Line: {i} Discarding POLYLINE with {len(vertices)} vertices which is degenerate")
                        continue

                    poly = PolyContour(vertices)
                    actual = poly.eWinding()
                    if actual == PolyContour.EWinding.UNKNOWN:
                        warnings.append(
                            f"Line: {i} Discarding POLYLINE with area 0 (degenerate) - defined with winding {PolyContour.strWindingAsString(declared)}"
                        )
                        continue
                    if actual != declared:
                        warnings.append(
                            f"Line: {i} POLYLINE defined with winding {PolyContour.strWindingAsString(declared)} actual winding is {PolyContour.strWindingAsString(actual)} (using actual)"
                        )
                    o_current_slice.AddContour(poly)
                    continue

                if line.startswith("$$GEOMETRYEND"):
                    b_geometry_ended = True
                    break

                if line.startswith("$$"):
                    warnings.append(f"Line: {i} Unsupported command {Utils.strShorten(line, 20)}")
                    line = ""
                    continue

                line = ""

        if o_current_slice is not None:
            slices.append(o_current_slice)

        result_stack = PolySliceStack(slices)
        return Cli.Result(
            oSlices=result_stack,
            oBBoxFile=((bbox_file[0][0], bbox_file[0][1], bbox_file[0][2]), (bbox_file[1][0], bbox_file[1][1], bbox_file[1][2])),
            bBinary=b_binary,
            fUnitsHeader=f_units_header,
            b32BitAlign=b_32bit_align,
            nVersion=n_version,
            strHeaderDate=str_header_date,
            nLayers=n_header_layer_numbers,
            strWarnings="\n".join(warnings),
        )


class CliIo(Cli):
    @staticmethod
    def WriteSlicesToCliFile(
        oSlices: PolySliceStack | Sequence[PolySlice],
        strFilePath: str | Path,
        eFormat: "Cli.EFormat" = Cli.EFormat.FirstLayerWithContent,
        strDate: str = "",
        fUnitsInMM: float = 0.0,
    ) -> None:
        Cli.WriteSlicesToCliFile(oSlices, strFilePath, eFormat, strDate, fUnitsInMM)

    @staticmethod
    def oSlicesFromCliFile(strFilePath: str | Path) -> "Cli.Result":
        return Cli.oSlicesFromCliFile(strFilePath)




from __future__ import annotations

from collections.abc import Callable, Sequence
from dataclasses import dataclass
from enum import Enum

import numpy as np
from picogk import Image

from .._types import Vector3Like, as_np3, clamp01

ImageLike = Image


class LineModulation:
    class ECoord(Enum):
        X = "X"
        Y = "Y"
        Z = "Z"

    RatioFunc = Callable[[float], float]

    def __init__(self, source: float | RatioFunc) -> None:
        self.m_fConstValue: float = 0.0
        self.m_aXValues: list[float] = []
        self.m_aYValues: list[float] = []

        if callable(source):
            self.m_oFunc = source
            return

        self.m_fConstValue = float(source)
        self.m_oFunc = self._const_line_func

    @classmethod
    def from_discrete(
        cls,
        aDiscretePoints: Sequence[Vector3Like],
        eValues: "LineModulation.ECoord",
        eAxis: "LineModulation.ECoord",
    ) -> "LineModulation":
        points = [as_np3(p) for p in aDiscretePoints]
        if not points:
            raise ValueError("aDiscretePoints cannot be empty")

        mod = cls(0.0)
        mod.m_aXValues = []
        mod.m_aYValues = []

        first_x = cls._coord(points[0], eAxis)
        first_y = cls._coord(points[0], eValues)
        mod.m_aXValues.append(first_x)
        mod.m_aYValues.append(first_y)

        for point in points[1:]:
            x_val = cls._coord(point, eAxis)
            y_val = cls._coord(point, eValues)
            if x_val > mod.m_aXValues[-1]:
                mod.m_aXValues.append(x_val)
                mod.m_aYValues.append(y_val)

        if mod.m_aXValues[0] > 0.0:
            mod.m_aXValues.insert(0, 0.0)
            mod.m_aYValues.insert(0, mod.m_aYValues[0])
        if mod.m_aXValues[-1] < 1.0:
            mod.m_aXValues.append(1.0)
            mod.m_aYValues.append(mod.m_aYValues[-1])

        mod.m_oFunc = mod._discrete_interp
        return mod

    @staticmethod
    def _coord(vecPt: np.ndarray, coord: "LineModulation.ECoord") -> float:
        if coord == LineModulation.ECoord.Y:
            return float(vecPt[1])
        if coord == LineModulation.ECoord.Z:
            return float(vecPt[2])
        return float(vecPt[0])

    def _const_line_func(self, fRatio: float) -> float:
        return self.m_fConstValue

    def _discrete_interp(self, fX: float) -> float:
        x = clamp01(fX)
        idx = int(np.searchsorted(self.m_aXValues, x, side="left"))
        upper = min(idx, len(self.m_aXValues) - 1)
        lower = max(upper - 1, 0)

        lower_x = self.m_aXValues[lower]
        upper_x = self.m_aXValues[upper]
        if upper_x == lower_x:
            return self.m_aYValues[lower]

        ratio = (x - lower_x) / (upper_x - lower_x)
        return self.m_aYValues[lower] + ratio * (self.m_aYValues[upper] - self.m_aYValues[lower])

    def fGetModulation(self, fRatio: float) -> float:
        return float(self.m_oFunc(float(fRatio)))

    def f_get_modulation(self, fRatio: float) -> float:
        return self.fGetModulation(fRatio)

    def __mul__(self, fFactor: float) -> "LineModulation":
        return LineModulation(lambda lr: float(fFactor) * self.fGetModulation(lr))

    def __rmul__(self, fFactor: float) -> "LineModulation":
        return self.__mul__(fFactor)

    def __add__(self, other: "LineModulation") -> "LineModulation":
        return LineModulation(lambda lr: self.fGetModulation(lr) + other.fGetModulation(lr))

    def __sub__(self, other: "LineModulation") -> "LineModulation":
        return LineModulation(lambda lr: self.fGetModulation(lr) - other.fGetModulation(lr))


@dataclass
class Distribution:
    m_fTotalLength: float
    m_oModulation: LineModulation


class GenericContour(Distribution):
    pass




from __future__ import annotations

import ctypes
import threading
from typing import TYPE_CHECKING
from _common.types import Vector3Like

from .._config import STRING_LENGTH
from .._native import native
from .._types import Vec3, as_vec3, vec3_tuple
# from .._viewer_protocol import IViewer as VedoViewer
if TYPE_CHECKING:
    from .._viewer import VedoViewer

class Library:
    _lock = threading.Lock()
    _running = False
    _app_exit = False
    _continue_task = True
    _viewer: VedoViewer | None = None
    voxel_size_mm: float = 0.5

    @staticmethod
    def _lib() -> ctypes.CDLL:
        return native().lib

    @classmethod
    def init(cls, voxel_size_mm: float) -> None:
        with cls._lock:
            if cls._running:
                raise RuntimeError("PicoGK is already running")
            cls._lib().Library_Init(ctypes.c_float(voxel_size_mm))
            cls._running = True
            cls._app_exit = False
            cls._continue_task = True
            cls.voxel_size_mm = voxel_size_mm

    @classmethod
    def destroy(cls) -> None:
        with cls._lock:
            if not cls._running:
                return
            cls._app_exit = True
            cls._lib().Library_Destroy()
            cls._running = False

    @classmethod
    def _string_fn(cls, fn_name: str) -> str:
        buffer = ctypes.create_string_buffer(STRING_LENGTH)
        getattr(cls._lib(), fn_name)(buffer)
        return buffer.value.decode("utf-8")

    @classmethod
    def name(cls) -> str:
        return cls._string_fn("Library_GetName")

    strName = name

    @classmethod
    def version(cls) -> str:
        return cls._string_fn("Library_GetVersion")

    strVersion = version

    @classmethod
    def build_info(cls) -> str:
        return cls._string_fn("Library_GetBuildInfo")

    strBuildInfo = build_info

    @classmethod
    def voxels_to_mm(cls, xyz: Vector3Like) -> tuple[float, float, float]:
        src = as_vec3(xyz)
        dst = Vec3()
        cls._lib().Library_VoxelsToMm(ctypes.byref(src), ctypes.byref(dst))
        return vec3_tuple(dst)

    vecVoxelsToMm = voxels_to_mm

    @classmethod
    def mm_to_voxels(cls, xyz: Vector3Like) -> tuple[float, float, float]:
        src = as_vec3(xyz)
        dst = Vec3()
        cls._lib().Library_MmToVoxels(ctypes.byref(src), ctypes.byref(dst))
        return vec3_tuple(dst)

    MmToVoxels = mm_to_voxels

    @classmethod
    def bContinueTask(cls, app_exit_only: bool = False) -> bool:
        return (not cls._app_exit) and (app_exit_only or cls._continue_task)

    @classmethod
    def EndTask(cls) -> None:
        cls._continue_task = False

    @classmethod
    def CancelEndTaskRequest(cls) -> None:
        cls._continue_task = True

    @classmethod
    def Log(cls, message: str, *args: object) -> None:
        if args:
            message = message.format(*args)
        print(message)

    @classmethod
    def SetViewer(cls, viewer: VedoViewer | None) -> None:
        cls._viewer = viewer

    @classmethod
    def oViewer(cls) -> VedoViewer | None:
        return cls._viewer

    @classmethod
    def ClearViewer(cls) -> None:
        cls._viewer = None

    @classmethod
    def Dispose(cls) -> None:
        cls.destroy()






from __future__ import annotations

from _common.types import ColorLike, Vector3Like, as_np3, as_tuple3, clamp01, normalized, np_vec3

Vec3 = tuple[float, float, float]


def as_vec3(value: Vector3Like) -> Vec3:
    return as_tuple3(value)


__all__ = ["ColorLike", "Vec3", "Vector3Like", "as_np3", "as_vec3", "clamp01", "normalized", "np_vec3"]



from __future__ import annotations

import ctypes
from typing import Optional

from ._errors import PicoGKDisposedError


class HandleOwner:
    """Base class for wrappers that own a native handle."""

    _destroy_func_name: Optional[str] = None

    def __init__(self, handle: int | ctypes.c_void_p, *, owns_handle: bool = True) -> None:
        self._handle = ctypes.c_void_p(int(handle) if not isinstance(handle, ctypes.c_void_p) else handle.value)
        self._owns_handle = owns_handle
        self._closed = False

    @property
    def handle(self) -> ctypes.c_void_p:
        self._ensure_open()
        return self._handle

    @property
    def is_closed(self) -> bool:
        return self._closed or not bool(self._handle.value)

    def _ensure_open(self) -> None:
        if self.is_closed:
            raise PicoGKDisposedError(f"{self.__class__.__name__} is closed")

    def _destroy_native(self) -> None:
        raise NotImplementedError

    def close(self) -> None:
        if self.is_closed:
            self._closed = True
            return
        try:
            if self._owns_handle:
                self._destroy_native()
        finally:
            self._handle = ctypes.c_void_p()
            self._closed = True

    def __enter__(self):
        self._ensure_open()
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        self.close()

    def __del__(self) -> None:
        try:
            self.close()
        except Exception:
            pass



from __future__ import annotations

from shape_kernel._types import Vector3Like

from lattice_library import IBeamThickness

from .coordinate_trafo import ICoordinateTrafo
from picogk import IImplicit
from .raw_tpms_patterns import IRawTPMSPattern
from .splitting_logic import ISplittingLogic


class ImplicitModular(IImplicit):
	def __init__(
		self,
		xRawTPMSPattern: IRawTPMSPattern,
		xWallThickness: IBeamThickness,
		xCoordinateTrafo: ICoordinateTrafo,
		xSplittingLogic: ISplittingLogic,
	) -> None:
		self.m_xWallThickness = xWallThickness
		self.m_xCoordinateTrafo = xCoordinateTrafo
		self.m_xRawTPMSPattern = xRawTPMSPattern
		self.m_xSplittingLogic = xSplittingLogic

	def fSignedDistance(self, vecPt: Vector3Like) -> float:
		fX, fY, fZ = self.m_xCoordinateTrafo.Apply(vecPt)
		fRawSignedDistance = self.m_xRawTPMSPattern.fGetSignedDistance(fX, fY, fZ)
		fWallThickness = self.m_xWallThickness.fGetBeamThickness(vecPt)
		return self.m_xSplittingLogic.fGetAdvancedSignedDistance(fRawSignedDistance, fWallThickness)


__all__ = ["ImplicitModular"]


from __future__ import annotations

import math

from picogk import Library
from shape_kernel import Cp, Sh, VecOperations
from shape_kernel._types import Vec3, as_np3, as_vec3

from .rhombic_tile import ISubDTile, LargeRhombicTile
from .robinson_triangle import RobinsonTriangle


class PenrosePattern:
    def __init__(self, nGenerations: int, fInitialSide: float = 20.0) -> None:
        self.m_nGenerations = int(nGenerations)
        self.m_aTileGenerations: list[list[ISubDTile]] = [[] for _ in range(self.m_nGenerations)]
        self.m_fInitialSide = float(fInitialSide)
        self.m_aTileGenerations[0] = self.aGetDefaultInitialTiles()

        for i in range(1, self.m_nGenerations):
            a_inflated_sub_tiles: list[ISubDTile] = []
            for x_tile in self.m_aTileGenerations[i - 1]:
                a_inflated_sub_tiles.extend(x_tile.aGetInflatedSubTiles())

            a_inflated_sub_tiles = self.aGetDeduplicatedSubTiles(a_inflated_sub_tiles)
            self.m_aTileGenerations[i] = a_inflated_sub_tiles

    def aGetDeduplicatedSubTiles(self, aSubTiles: list[ISubDTile]) -> list[ISubDTile]:
        a_ref_centres: list[Vec3] = []
        a_deduplicated_sub_tiles: list[ISubDTile] = []

        for x_sub_tile in aSubTiles:
            a_vertices = x_sub_tile.aGetTileVertices()
            vec_centre = sum((as_np3(v) for v in a_vertices), start=as_np3((0.0, 0.0, 0.0))) / float(len(a_vertices))
            vec_rounded_centre = as_vec3((round(float(vec_centre[0]), 2), round(float(vec_centre[1]), 2), round(float(vec_centre[2]), 2)))

            if vec_rounded_centre not in a_ref_centres:
                a_ref_centres.append(vec_rounded_centre)
                a_deduplicated_sub_tiles.append(x_sub_tile)

        return a_deduplicated_sub_tiles

    def aGetDefaultInitialTiles(self) -> list[ISubDTile]:
        a_initial_tiles: list[ISubDTile] = []
        f_theta = 108.0 / 180.0 * math.pi
        n_symmetry = 5

        for i in range(n_symmetry):
            f_alpha = (2.0 * math.pi) / float(n_symmetry) * float(i)
            vec_a: Vec3 = (0.0, 0.0, 0.0)
            vec_b = VecOperations.vecRotateAroundZ((self.m_fInitialSide, 0.0, 0.0), f_alpha)
            vec_c = VecOperations.vecRotateAroundZ(vec_a, f_theta, vec_b)

            o_tri = RobinsonTriangle(vec_a, vec_b, vec_c)
            o_tile: ISubDTile = LargeRhombicTile(o_tri)
            a_initial_tiles.append(o_tile)

        return a_initial_tiles

    def PreviewGeneration(self, nGen: int) -> None:
        try:
            a_tiles = self.m_aTileGenerations[int(nGen)]
            Library.Log(f"Number of Tiles = {len(a_tiles)}.")
            for x_tile in a_tiles:
                a_vertices = x_tile.aGetTileVertices()
                a_vertices.append(a_vertices[0])
                Sh.PreviewLine(a_vertices, Cp.clrBlack)
        except Exception as exc:
            raise Exception("Generation not found.") from exc


__all__ = ["PenrosePattern"]



from __future__ import annotations

from abc import ABC, abstractmethod


class ISplittingLogic(ABC):
	@abstractmethod
	def fGetAdvancedSignedDistance(self, fSignedDistance: float, fWallThickness: float) -> float:
		raise NotImplementedError


class FullWallLogic(ISplittingLogic):
	def fGetAdvancedSignedDistance(self, fSignedDistance: float, fWallThickness: float) -> float:
		return abs(float(fSignedDistance)) - 0.5 * float(fWallThickness)


class FullVoidLogic(ISplittingLogic):
	def fGetAdvancedSignedDistance(self, fSignedDistance: float, fWallThickness: float) -> float:
		return -(abs(float(fSignedDistance)) - 0.5 * float(fWallThickness))


class PositiveHalfWallLogic(ISplittingLogic):
	def fGetAdvancedSignedDistance(self, fSignedDistance: float, fWallThickness: float) -> float:
		fSD = float(fSignedDistance)
		return max(fSD, abs(fSD) - 0.5 * float(fWallThickness))


class NegativeHalfWallLogic(ISplittingLogic):
	def fGetAdvancedSignedDistance(self, fSignedDistance: float, fWallThickness: float) -> float:
		fSD = float(fSignedDistance)
		return max(-fSD, abs(fSD) - 0.5 * float(fWallThickness))


class PositiveVoidLogic(ISplittingLogic):
	def fGetAdvancedSignedDistance(self, fSignedDistance: float, fWallThickness: float) -> float:
		fFinalDist = max(0.0, float(fSignedDistance)) - 0.5 * float(fWallThickness)
		return -fFinalDist


class NegativeVoidLogic(ISplittingLogic):
	def fGetAdvancedSignedDistance(self, fSignedDistance: float, fWallThickness: float) -> float:
		fFinalDist = max(0.0, -float(fSignedDistance)) - 0.5 * float(fWallThickness)
		return -fFinalDist


__all__ = [
	"ISplittingLogic",
	"FullWallLogic",
	"FullVoidLogic",
	"PositiveHalfWallLogic",
	"NegativeHalfWallLogic",
	"PositiveVoidLogic",
	"NegativeVoidLogic",
]